# -*- coding: utf-8 -*-

from skyllh.datasets.i3 import (
    PublicData_10y_ps,
    PublicData_10y_ps_wMC,
    PublicData_10y_ps_wMCEq,
)

data_samples = {
    'PublicData_10y_ps': PublicData_10y_ps,
    'PublicData_10y_ps_wMC': PublicData_10y_ps_wMC,
    'PublicData_10y_ps_wMCEq': PublicData_10y_ps_wMCEq
}



# -*- coding: utf-8 -*-

import numpy as np

import scipy.interpolate

from skyllh.core import multiproc
from skyllh.core.binning import BinningDefinition
from skyllh.core.dataset import (
    Dataset,
    DatasetData
)
from skyllh.core.livetime import Livetime
from skyllh.core.parameters import ParameterGrid
from skyllh.core.detsigyield import (
    get_integrated_livetime_in_days
)
from skyllh.physics.flux import (
    PowerLawFlux,
    get_conversion_factor_to_internal_flux_unit
)
from skyllh.i3.detsigyield import (
    PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod,
    PowerLawFluxPointLikeSourceI3DetSigYield
)
from skyllh.analyses.i3.publicdata_ps.aeff import (
    load_effective_area_array
)


class PublicDataPowerLawFluxPointLikeSourceI3DetSigYieldImplMethod(
        PowerLawFluxPointLikeSourceI3DetSigYieldImplMethod,
        multiproc.IsParallelizable):
    """This detector signal yield constructor class constructs a
    detector signal yield instance for a variable power law flux model, which
    has the spectral index gamma as fit parameter, assuming a point-like source.
    It constructs a two-dimensional spline function in sin(dec) and gamma, using
    a :class:`scipy.interpolate.RectBivariateSpline`. Hence, the detector signal
    yield can vary with the declination and the spectral index, gamma, of the
    source.

    This detector signal yield implementation method works with a
    PowerLawFlux flux model.

    It is tailored to the IceCube detector at the South Pole, where the
    effective area depends solely on the zenith angle, and hence on the
    declination, of the source.

    It takes the effective area for the detector signal yield from the auxilary
    detector effective area data file given by the public data.
    """

    def __init__(
            self, gamma_grid, spline_order_sinDec=2, spline_order_gamma=2,
            ncpu=None):
        """Creates a new IceCube detector signal yield constructor instance for
        a power law flux model. It requires the effective area from the public
        data, and a gamma parameter grid to compute the gamma dependency of the
        detector signal yield.

        Parameters
        ----------
        gamma_grid : ParameterGrid instance
            The ParameterGrid instance which defines the grid of gamma values.
        spline_order_sinDec : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the sin(dec) axis.
            The default is 2.
        spline_order_gamma : int
            The order of the spline function for the logarithmic values of the
            detector signal yield along the gamma axis.
            The default is 2.
        ncpu : int | None
            The number of CPUs to utilize. Global setting will take place if
            not specified, i.e. set to None.
        """
        super().__init__(
            gamma_grid=gamma_grid,
            sin_dec_binning=None,
            spline_order_sinDec=spline_order_sinDec,
            spline_order_gamma=spline_order_gamma,
            ncpu=ncpu)

    def construct_detsigyield(
            self, dataset, data, fluxmodel, livetime, ppbar=None):
        """Constructs a detector signal yield 2-dimensional log spline
        function for the given power law flux model with varying gamma values.

        Parameters
        ----------
        dataset : Dataset instance
            The Dataset instance holding the sin(dec) binning definition.
        data : DatasetData instance
            The DatasetData instance holding the monte-carlo event data.
            This implementation loads the effective area from the provided
            public data and hence does not need monte-carlo data.
        fluxmodel : FluxModel
            The flux model instance. Must be an instance of PowerLawFlux.
        livetime : float | Livetime instance
            The live-time in days or an instance of Livetime to use for the
            detector signal yield.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Returns
        -------
        detsigyield : PowerLawFluxPointLikeSourceI3DetSigYield instance
            The DetSigYield instance for a point-like source with a power law
            flux with variable gamma parameter.
        """
        # Check for the correct data types of the input arguments.
        if(not isinstance(dataset, Dataset)):
            raise TypeError('The dataset argument must be an instance of '
                            'Dataset!')
        if(not isinstance(data, DatasetData)):
            raise TypeError('The data argument must be an instance of '
                            'DatasetData!')
        if(not self.supports_fluxmodel(fluxmodel)):
            raise TypeError('The DetSigYieldImplMethod "%s" does not support '
                            'the flux model "%s"!' % (
                                self.__class__.__name__,
                                fluxmodel.__class__.__name__))
        if((not isinstance(livetime, float)) and
           (not isinstance(livetime, Livetime))):
            raise TypeError('The livetime argument must be an instance of '
                            'float or Livetime!')

        # Get integrated live-time in days.
        livetime_days = get_integrated_livetime_in_days(livetime)

        # Calculate conversion factor from the flux model unit into the internal
        # flux unit GeV^-1 cm^-2 s^-1.
        toGeVcm2s = get_conversion_factor_to_internal_flux_unit(fluxmodel)

        # Load the effective area data from the public dataset.
        aeff_fnames = dataset.get_abs_pathfilename_list(
            dataset.get_aux_data_definition('eff_area_datafile'))
        (
            aeff_arr,
            sin_true_dec_binedges_lower,
            sin_true_dec_binedges_upper,
            log_true_e_binedges_lower,
            log_true_e_binedges_upper
        ) = load_effective_area_array(aeff_fnames)

        # Calculate the detector signal yield in sin_dec vs gamma.
        def hist(
                energy_bin_edges_lower, energy_bin_edges_upper,
                aeff, fluxmodel):
            """Creates a histogram of the detector signal yield for the given
            sin(dec) binning.

            Parameters
            ----------
            energy_bin_edges_lower : 1d ndarray
                The array holding the lower bin edges in E_nu/GeV.
            energy_bin_edges_upper : 1d ndarray
                The array holding the upper bin edges in E_nu/GeV.
            aeff : (n_bins_sin_dec, n_bins_log_energy)-shaped 2d ndarray
                The effective area binned data array.

            Returns
            -------
            h : (n_bins_sin_dec,)-shaped 1d ndarray
                The numpy array containing the detector signal yield values for
                the different sin_dec bins and the given flux model.
            """
            # Create histogram for the number of neutrinos with each energy
            # bin.
            h_phi = fluxmodel.get_integral(
                energy_bin_edges_lower, energy_bin_edges_upper)

            # Sum over the enegry bins for each sin_dec row.
            h = np.sum(aeff*h_phi, axis=1)

            return h

        energy_bin_edges_lower = np.power(10, log_true_e_binedges_lower)
        energy_bin_edges_upper = np.power(10, log_true_e_binedges_upper)

        # Make a copy of the gamma grid and extend the grid by one bin on each
        # side.
        gamma_grid = self._gamma_grid.copy()
        gamma_grid.add_extra_lower_and_upper_bin()

        # Construct the arguments for the hist function to be used in the
        # multiproc.parallelize function.
        args_list = [
            ((energy_bin_edges_lower,
              energy_bin_edges_upper,
              aeff_arr,
              fluxmodel.copy({'gamma': gamma})), {})
            for gamma in gamma_grid.grid
        ]
        h = np.vstack(
            multiproc.parallelize(
                hist, args_list, self.ncpu, ppbar=ppbar)).T
        h *= toGeVcm2s * livetime_days * 86400.

        # Create a 2d spline in log of the detector signal yield.
        sin_dec_bincenters = 0.5*(
            sin_true_dec_binedges_lower + sin_true_dec_binedges_upper)
        log_spl_sinDec_gamma = scipy.interpolate.RectBivariateSpline(
            sin_dec_bincenters, gamma_grid.grid, np.log(h),
            kx=self.spline_order_sinDec, ky=self.spline_order_gamma, s=0)

        # Construct the detector signal yield instance with the created spline.
        sin_dec_binedges = np.concatenate(
            (sin_true_dec_binedges_lower, [sin_true_dec_binedges_upper[-1]]))
        sin_dec_binning = BinningDefinition('sin_dec', sin_dec_binedges)
        detsigyield = PowerLawFluxPointLikeSourceI3DetSigYield(
            self, dataset, fluxmodel, livetime, sin_dec_binning, log_spl_sinDec_gamma)

        return detsigyield



# -*- coding: utf-8 -*-

import numpy as np


def rotate_spherical_vector(ra1, dec1, ra2, dec2, ra3, dec3):
    """Calculates the rotation matrix R to rotate the spherical vector
    (ra1,dec1) onto the direction (ra2,dec2), and performs this rotation on the
    spherical vector (ra3,dec3).

    In practice (ra1,dec1) refers to the true location of a MC event,
    (ra2,dec2) the true location of the signal source, and (ra3,dec3) the
    reconstructed location of the MC event, which should get rotated according
    to the rotation of the two true directions.
    """
    # Make sure, the inputs are 1D arrays.
    ra1 = np.atleast_1d(ra1)
    dec1 = np.atleast_1d(dec1)
    ra2 = np.atleast_1d(ra2)
    dec2 = np.atleast_1d(dec2)
    ra3 = np.atleast_1d(ra3)
    dec3 = np.atleast_1d(dec3)

    assert (len(ra1) == len(dec1) ==
            len(ra2) == len(dec2) ==
            len(ra3) == len(dec3)
           ), 'All input argument arrays must be of the same length!'

    N_event = len(ra1)

    # Calculate the space angle alpha between vector 1 and vector 2, and
    # correct for possible rounding erros.
    cos_alpha = (np.cos(ra2 - ra1) * np.cos(dec1) * np.cos(dec2)
                 + np.sin(dec1) * np.sin(dec2))
    cos_alpha[cos_alpha >  1] =  1
    cos_alpha[cos_alpha < -1] = -1
    alpha = np.arccos(cos_alpha)

    # Define the three 3D-vectors in spherical coordinate system. Each vector
    # is a (N_event,3)-shaped 2D array.
    vec1 = np.vstack([np.cos(ra1) * np.cos(dec1),
                      np.sin(ra1) * np.cos(dec1),
                      np.sin(dec1)]).T
    vec2 = np.vstack([np.cos(ra2) * np.cos(dec2),
                      np.sin(ra2) * np.cos(dec2),
                      np.sin(dec2)]).T
    vec3 = np.vstack([np.cos(ra3) * np.cos(dec3),
                      np.sin(ra3) * np.cos(dec3),
                      np.sin(dec3)]).T

    # Calculate the normalized rotation axis vector, nrot. nrot is a
    # (N_event,3)-shaped ndarray.
    nrot = np.cross(vec1, vec2)
    norm = np.sqrt(np.sum(nrot**2, axis=1))
    nrot[norm > 0] /= norm[np.newaxis, norm > 0].T

    # Define the diagonal 3D unit matrix.
    one = np.diagflat(np.ones(3))

    # Calculate the rotation matrix R_i for each event i and perform the
    # rotation on vector 3 for each event.
    vec = np.empty((N_event,3), dtype=np.float64)

    sin_alpha = np.sin(alpha)
    twopi = 2*np.pi
    # Remap functions to avoid Python's (.)-resolution millions of times.
    (np_outer, np_dot, np_roll, np_diag, np_T) = (
     np.outer, np.dot, np.roll, np.diag, np.transpose)
    for i in range(N_event):
        cos_alpha_i = cos_alpha[i]
        nrot_i = nrot[i]
        nrotTnrot_i = np_outer(nrot_i, nrot_i)

        # Calculate cross product matrix, nrotx_i:
        # A[ij] = x_i * y_j - y_i * x_j
        skv = np_roll(np_roll(np_diag(nrot_i), shift=1, axis=1), shift=-1, axis=0)
        nrotx_i = skv - np_T(skv)

        # Calculate rotation matrix, R_i.
        R_i = ((1. - cos_alpha_i) * nrotTnrot_i
               + one*cos_alpha_i
               + sin_alpha[i] * nrotx_i)
        vec[i] = np_dot(R_i, np_T(vec3[i]))

    ra = np.arctan2(vec[:, 1], vec[:, 0])
    ra += np.where(ra < 0., twopi, 0.)
    dec = np.arcsin(vec[:, 2])

    return (ra, dec)



# -*- coding: utf-8 -*-

import numpy as np
import time

from skyllh.core import display
from skyllh.core.py import classname

"""The timing module provides code execution timing functionalities. The
TimeLord class keeps track of execution times of specific code segments,
called "tasks". The TaskTimer class can be used within a `with`
statement to time the execution of the code within the `with` block.
"""

class TaskRecord(object):
    def __init__(self, name, tstart, tend):
        """Creates a new TaskRecord instance.

        Parameters
        ----------
        name : str
            The name of the task.
        tstart : float | 1d ndarray of float
            The start time(s) of the task in seconds.
        tend : float | 1d ndarray of float
            The end time(s) of the task in seconds.
        """
        self.name = name

        tstart = np.atleast_1d(tstart)
        tend = np.atleast_1d(tend)

        if(len(tstart) != len(tend)):
            raise ValueError('The number of start and end time stamps must '
                'be equal!')

        self._tstart_list = list(tstart)
        self._tend_list = list(tend)

    @property
    def tstart(self):
        """(read-only) The time stamps the execution of this task started.
        """
        return self._tstart_list

    @property
    def tend(self):
        """(read-only) The time stamps the execution of this task was stopped.
        """
        return self._tend_list

    @property
    def duration(self):
        """(read-only) The total duration (without time overlap) the task was
        executed.
        """
        # Create a (2,Niter)-shaped 2D ndarray holding the start and end time
        # stamps of the task executions. This array gets then sorted by the
        # start time stamps.
        arr = np.sort(np.vstack((self._tstart_list, self._tend_list)), axis=1)

        d = arr[1,0] - arr[0,0]
        last_tend = arr[1,0]
        for idx in range(1, arr.shape[1]):
            tstart = arr[0,idx]
            tend = arr[1,idx]
            if(tend <= last_tend):
                continue
            if(tstart <= last_tend and tend > last_tend):
                d += tend - last_tend
            elif(tstart >= last_tend):
                d += tend - tstart
            last_tend = tend

        return d

    @property
    def niter(self):
        """(read-only) The number of times this task was executed.
        """
        return len(self._tstart_list)

    def join(self, tr):
        """Joins this TaskRecord with the given TaskRecord instance.

        Parameters
        ----------
        tr : TaskRecord
            The instance of TaskRecord that should be joined with this
            TaskRecord instance.
        """
        self._tstart_list.extend(tr._tstart_list)
        self._tend_list.extend(tr._tend_list)


class TimeLord(object):
    def __init__(self):
        self._task_records = []
        self._task_records_name_idx_map = {}

    @property
    def task_name_list(self):
        """(read-only) The list of task names.
        """
        return list(self._task_records_name_idx_map.keys())

    def add_task_record(self, tr):
        """Adds a given task record to the internal list of task records.
        """
        tname = tr.name

        if(self.has_task_record(tname)):
            # The TaskRecord already exists. Update the task record.
            self_tr = self.get_task_record(tname)
            self_tr.join(tr)
            return

        self._task_records.append(tr)
        self._task_records_name_idx_map[tr.name] = len(self._task_records)-1

    def get_task_record(self, name):
        """Retrieves a task record of the given name.

        Parameters
        ----------
        name : str
            The name of the task record.

        Returns
        -------
        task_record : TaskRecord
            The instance of TaskRecord with the requested name.
        """
        return self._task_records[self._task_records_name_idx_map[name]]

    def has_task_record(self, name):
        """Checks if this TimeLord instance has a task record of the given name.

        Parameters
        ----------
        name : str
            The name of the task record.

        Returns
        -------
        check : bool
            ``True`` if this TimeLord instance has a task record of the given
            name, and ``False`` otherwise.
        """
        return name in self._task_records_name_idx_map

    def join(self, tl):
        """Joins a given TimeLord instance with this TimeLord instance. Tasks
        of the same name will be updated and new tasks will be added.

        Parameters
        ----------
        tl : TimeLord instance
            The instance of TimeLord whos tasks should be joined with the tasks
            of this TimeLord instance.
        """
        for tname in tl.task_name_list:
            other_tr = tl.get_task_record(tname)
            if(self.has_task_record(tname)):
                # Update the task record.
                tr = self.get_task_record(tname)
                tr.join(other_tr)
            else:
                # Add a new task record.
                self.add_task_record(other_tr)

    def task_timer(self, name):
        """Creates TaskTimer instance for the given task name.
        """
        return TaskTimer(self, name)

    def __str__(self):
        """Generates a pretty string for this time lord.
        """
        task_name_list = self.task_name_list
        n_tasks = len(task_name_list)

        s = f'{classname(self)}: Executed tasks:'
        if(n_tasks == 0):
            s += ' None.'
            return s

        task_name_len_list = [ len(task_name) for task_name in task_name_list ]
        max_task_name_len = np.minimum(
            np.max(task_name_len_list), display.PAGE_WIDTH-25)

        for i in range(n_tasks):
            tr = self._task_records[i]
            task_name = tr.name[0:max_task_name_len]
            t = tr.duration / tr.niter
            line = '\n[{task_name:'+str(max_task_name_len)+'s}] {t:7.{p}{c}} '\
                'sec/iter ({niter:d})'
            s += line.format(
                task_name=task_name,
                t=t,
                p=1 if t>1e3 or t<1e-3 else 3,
                c='e' if t > 1e3 or t < 1e-3 else 'f',
                niter=tr.niter)

        return s


class TaskTimer(object):
    def __init__(self, time_lord, name):
        """
        Parameters
        ----------
        time_lord : instance of TimeLord
            The TimeLord instance that keeps track of the recorded tasks.
        name : str
            The name of the task.
        """
        self.time_lord = time_lord
        self.name = name

        self._start = None
        self._end = None

    @property
    def time_lord(self):
        """The TimeLord instance that keeps track of the recorded tasks. This
        can be None, which means that the task should not get recorded.
        """
        return self._time_lord
    @time_lord.setter
    def time_lord(self, lord):
        if(lord is not None):
            if(not isinstance(lord, TimeLord)):
                raise TypeError('The time_lord property must be None or an '
                    'instance of TimeLord!')
        self._time_lord = lord

    @property
    def name(self):
        """The name if the task.
        """
        return self._name
    @name.setter
    def name(self, name):
        if(not isinstance(name, str)):
            raise TypeError('The name property must be an instance of str!')
        self._name = name

    @property
    def duration(self):
        """The duration in seconds the task was executed.
        """
        return (self._end - self._start)

    def __enter__(self):
        """This gets executed when entering the `with` block.
        """
        self._start = time.process_time()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """This gets executed when exiting the `with` block.
        """
        self._end = time.process_time()

        if(self._time_lord is None):
            return

        self._time_lord.add_task_record(TaskRecord(
            self._name, self._start, self._end))



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.livetime import Livetime
from skyllh.core import storage
from skyllh.i3.dataset import I3Dataset

class I3Livetime(Livetime):
    """The I3Livetime class provides the functionality to load a Livetime object
    from a good-run-list data file.
    """
    @staticmethod
    def from_GRL_files(pathfilenames):
        """Loads an I3Livetime instance from the given good-run-list (GRL) data
        file. The data file needs to contain the following data fields:

            - start : float
                The MJD of the run start.
            - stop : float
                The MJD of the run stop.

        Parameters
        ----------
        pathfilenames : str | list of str
            The list of fully qualified file names of the GRL data files.

        Returns
        -------
        livetime : I3Livetime instance
            The created I3Livetime instance for the provided GRL data.
        """
        grl_data = storage.create_FileLoader(pathfilenames).load_data()

        uptime_mjd_intervals_arr = np.hstack((
            grl_data['start'].reshape((len(grl_data),1)),
            grl_data['stop'].reshape((len(grl_data),1))
        ))

        return I3Livetime(uptime_mjd_intervals_arr)

    @staticmethod
    def from_I3Dataset(ds):
        """Loads an I3Livetime instance from a given I3Dataset instance, which
        must have a good-run-list (GRL) files defined.

        Parameters
        ----------
        ds : I3Dataset instance
            The instance of I3Dataset which defined the good-run-list (GRL)
            files for the dataset.

        Returns
        -------
        livetime : I3Livetime instance
            The created I3Livetime instance for the GRL data from the provided
            dataset.
        """
        if(not isinstance(ds, I3Dataset)):
            raise TypeError('The ds argument must be an instance of I3Dataset!')
        if(len(ds.grl_pathfilename_list) == 0):
            raise ValueError('No GRL files have been defined for the given dataset!')
        return I3Livetime.from_GRL_files(ds.grl_pathfilename_list)

    def __init__(self, uptime_mjd_intervals_arr):
        super(I3Livetime, self).__init__(uptime_mjd_intervals_arr)



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.binning import (
    BinningDefinition,
    UsesBinning,
)
from skyllh.core.pdf import (
    EnergyPDF,
    IsBackgroundPDF,
    PDFAxis
)
from skyllh.core.storage import DataFieldRecordArray
from skyllh.core.timing import TaskTimer
from skyllh.core.smoothing import (
    UNSMOOTH_AXIS,
    SmoothingFilter,
    HistSmoothingMethod,
    NoHistSmoothingMethod,
    NeighboringBinHistSmoothingMethod
)
from skyllh.core.timing import TaskTimer

from scipy.stats import gaussian_kde


class PDEnergyPDF(EnergyPDF, UsesBinning):
    """This is the base class for IceCube specific energy PDF models.
    IceCube energy PDFs depend solely on the energy and the
    zenith angle, and hence, on the declination of the event.

    The IceCube energy PDF is modeled as a 1d histogram in energy,
    but for different sin(declination) bins, hence, stored as a 2d histogram.
    """

    _KDE_BW_NORTH = 0.4
    _KDE_BW_SOUTH = 0.32

    def __init__(self, data_logE, data_sinDec, data_mcweight, data_physicsweight,
                 logE_binning, sinDec_binning, smoothing_filter, kde_smoothing=False):
        """Creates a new IceCube energy PDF object.

        Parameters
        ----------
        data_logE : 1d ndarray
            The array holding the log10(E) values of the events.
        data_sinDec : 1d ndarray
            The array holding the sin(dec) values of the events.
        data_mcweight : 1d ndarray
            The array holding the monte-carlo weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        data_physicsweight : 1d ndarray
            The array holding the physics weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        logE_binning : BinningDefinition
            The binning definition for the log(E) axis.
        sinDec_binning : BinningDefinition
            The binning definition for the sin(declination) axis.
        smoothing_filter : SmoothingFilter instance | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        kde_smoothing : bool
            Apply a kde smoothing to the energy pdf for each bin in sin(dec).
            This is useful for signal injections, because it ensures that the
            background is not zero when injecting high energy events.
            Default: False.
        """
        super(PDEnergyPDF, self).__init__()

        # Define the PDF axes.
        self.add_axis(PDFAxis(name='log_energy',
                              vmin=logE_binning.lower_edge,
                              vmax=logE_binning.upper_edge))
        self.add_axis(PDFAxis(name='sin_dec',
                              vmin=sinDec_binning.lower_edge,
                              vmax=sinDec_binning.upper_edge))

        self.add_binning(logE_binning, 'log_energy')
        self.add_binning(sinDec_binning, 'sin_dec')

        # Create the smoothing method instance tailored to the energy PDF.
        # We will smooth only the first axis (logE).
        if((smoothing_filter is not None) and
           (not isinstance(smoothing_filter, SmoothingFilter))):
            raise TypeError(
                'The smoothing_filter argument must be None or an instance of SmoothingFilter!')
        if(smoothing_filter is None):
            self.hist_smoothing_method = NoHistSmoothingMethod()
        else:
            self.hist_smoothing_method = NeighboringBinHistSmoothingMethod(
                (smoothing_filter.axis_kernel_array, UNSMOOTH_AXIS))

        # We have to figure out, which histogram bins are zero due to no
        # monte-carlo coverage, and which due to zero physics model
        # contribution.

        # Create a 2D histogram with only the MC events to determine the MC
        # coverage.
        (h, bins_logE, bins_sinDec) = np.histogram2d(
            data_logE, data_sinDec,
            bins=[
                logE_binning.binedges, sinDec_binning.binedges],
            range=[
                logE_binning.range, sinDec_binning.range],
            density=False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered = h > 0

        # Select the events which have MC coverage but zero physics
        # contribution, i.e. the physics model predicts zero contribution.
        mask = data_physicsweight == 0.

        # Create a 2D histogram with only the MC events that have zero physics
        # contribution. Note: By construction the zero physics contribution bins
        # are a subset of the MC covered bins.
        (h, bins_logE, bins_sinDec) = np.histogram2d(
            data_logE[mask], data_sinDec[mask],
            bins=[
                logE_binning.binedges, sinDec_binning.binedges],
            range=[
                logE_binning.range, sinDec_binning.range],
            density=False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered_zero_physics = h > 0

        if kde_smoothing:
            # If a bandwidth is passed, apply a KDE-based smoothing with the given
            # bw parameter as bandwidth for the fit.
            if not isinstance(kde_smoothing, bool):
                raise ValueError(
                    "The bandwidth parameter must be True or False!")
            kde_pdf = np.empty(
                (len(sinDec_binning.bincenters),), dtype=object)
            data_logE_masked = data_logE[~mask]
            data_sinDec_masked = data_sinDec[~mask]
            for i in range(len(sinDec_binning.bincenters)):
                sindec_mask = np.logical_and(
                    data_sinDec_masked >= sinDec_binning.binedges[i],
                    data_sinDec_masked < sinDec_binning.binedges[i+1]
                )
                this_energy = data_logE_masked[sindec_mask]
                if sinDec_binning.binedges[i] >= 0:
                    kde_pdf[i] = gaussian_kde(
                        this_energy, bw_method=self._KDE_BW_NORTH)
                else:
                    kde_pdf[i] = gaussian_kde(
                        this_energy, bw_method=self._KDE_BW_SOUTH)
            h = np.vstack(
                [kde_pdf[i].evaluate(logE_binning.bincenters)
                 for i in range(len(sinDec_binning.bincenters))]).T

        else:
            # Create a 2D histogram with only the data which has physics
            # contribution. We will do the normalization along the logE
            # axis manually.
            data_weights = data_mcweight[~mask] * data_physicsweight[~mask]
            (h, bins_logE, bins_sinDec) = np.histogram2d(
                data_logE[~mask], data_sinDec[~mask],
                bins=[
                    logE_binning.binedges, sinDec_binning.binedges],
                weights=data_weights,
                range=[
                    logE_binning.range, sinDec_binning.range],
                density=False)

        # Calculate the normalization for each logE bin. Hence we need to sum
        # over the logE bins (axis 0) for each sin(dec) bin and need to divide
        # by the logE bin widths along the sin(dec) bins. The result array norm
        # is a 2D array of the same shape as h.
        norms = np.sum(h, axis=(0,))[np.newaxis, ...] * \
            np.diff(logE_binning.binedges)[..., np.newaxis]
        h /= norms
        h = self._hist_smoothing_method.smooth(h)

        self._hist_logE_sinDec = h

    @property
    def hist_smoothing_method(self):
        """The HistSmoothingMethod instance defining the smoothing filter of the
        energy PDF histogram.
        """
        return self._hist_smoothing_method

    @hist_smoothing_method.setter
    def hist_smoothing_method(self, method):
        if(not isinstance(method, HistSmoothingMethod)):
            raise TypeError(
                'The hist_smoothing_method property must be an instance of HistSmoothingMethod!')
        self._hist_smoothing_method = method

    @property
    def hist(self):
        """(read-only) The 2D logE-sinDec histogram array.
        """
        return self._hist_logE_sinDec

    @property
    def hist_mask_mc_covered(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage.
        """
        return self._hist_mask_mc_covered

    @property
    def hist_mask_mc_covered_zero_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage but zero physics
        contribution.
        """
        return self._hist_mask_mc_covered_zero_physics

    @property
    def hist_mask_mc_covered_with_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage and has physics
        contribution.
        """
        return self._hist_mask_mc_covered & ~self._hist_mask_mc_covered_zero_physics

    def get_prob(self, tdm, fitparams=None, tl=None):
        """Calculates the energy probability (in logE) of each event.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the data events for which the
            probability should be calculated for. The following data fields must
            exist:

            - 'log_energy' : float
                The logarithm of the energy value of the event.
            - 'sin_dec' : float
                The sin(declination) value of the event.

        fitparams : None
            Unused interface parameter.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        prob : 1D (N_events,) shaped ndarray
            The array with the energy probability for each event.
        """
        get_data = tdm.get_data

        logE_binning = self.get_binning('log_energy')
        sinDec_binning = self.get_binning('sin_dec')

        logE_idx = np.digitize(
            get_data('log_energy'), logE_binning.binedges) - 1
        sinDec_idx = np.digitize(
            get_data('sin_dec'), sinDec_binning.binedges) - 1

        with TaskTimer(tl, 'Evaluating logE-sinDec histogram.'):
            prob = self._hist_logE_sinDec[(logE_idx, sinDec_idx)]

        return prob


class PDDataBackgroundI3EnergyPDF(PDEnergyPDF, IsBackgroundPDF):
    """This is the IceCube energy background PDF, which gets constructed from
    experimental data. This class is derived from I3EnergyPDF.
    """

    def __init__(self, data_exp, logE_binning, sinDec_binning,
                 smoothing_filter=None, kde_smoothing=False):
        """Constructs a new IceCube energy background PDF from experimental
        data.

        Parameters
        ----------
        data_exp : instance of DataFieldRecordArray
            The array holding the experimental data. The following data fields
            must exist:

            - 'log_energy' : float
                The logarithm of the reconstructed energy value of the data
                event.
            - 'sin_dec' : float
                The sine of the reconstructed declination of the data event.

        logE_binning : BinningDefinition
            The binning definition for the binning in log10(E).
        sinDec_binning : BinningDefinition
            The binning definition for the sin(declination).
        smoothing_filter : SmoothingFilter instance | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        """
        if(not isinstance(data_exp, DataFieldRecordArray)):
            raise TypeError('The data_exp argument must be an instance of '
                            'DataFieldRecordArray!')

        data_logE = data_exp['log_energy']
        data_sinDec = data_exp['sin_dec']
        # For experimental data, the MC and physics weight are unity.
        data_mcweight = np.ones((len(data_exp),))
        data_physicsweight = data_mcweight

        # Create the PDF using the base class.
        super(PDDataBackgroundI3EnergyPDF, self).__init__(
            data_logE, data_sinDec, data_mcweight, data_physicsweight,
            logE_binning, sinDec_binning, smoothing_filter, kde_smoothing
        )


class PDMCBackgroundI3EnergyPDF(EnergyPDF, IsBackgroundPDF, UsesBinning):
    """This class provides a background energy PDF constructed from the public
    data and a monte-carlo background flux model.
    """

    def __init__(
            self, pdf_log10emu_sindecmu, log10emu_binning, sindecmu_binning,
            **kwargs):
        """Constructs a new background energy PDF with the given PDF data and
        binning.

        Parameters
        ----------
        pdf_log10emu_sindecmu : 2D numpy ndarray
            The (n_log10emu, n_sindecmu)-shaped 2D numpy ndarray holding the
            PDF values in unit 1/log10(E_mu/GeV).
            A copy of this data will be created and held within this class
            instance.
        log10emu_binning : BinningDefinition
            The binning definition for the binning in log10(E_mu/GeV).
        sindecmu_binning : BinningDefinition
            The binning definition for the binning in sin(dec_mu).
        """
        if not isinstance(pdf_log10emu_sindecmu, np.ndarray):
            raise TypeError(
                'The pdf_log10emu_sindecmu argument must be an instance of '
                'numpy.ndarray!')
        if not isinstance(sindecmu_binning, BinningDefinition):
            raise TypeError(
                'The sindecmu_binning argument must be an instance of '
                'BinningDefinition!')
        if not isinstance(log10emu_binning, BinningDefinition):
            raise TypeError(
                'The log10emu_binning argument must be an instance of '
                'BinningDefinition!')

        super().__init__(**kwargs)

        self.add_axis(PDFAxis(
            log10emu_binning.name,
            log10emu_binning.lower_edge,
            log10emu_binning.upper_edge,
        ))

        self.add_axis(PDFAxis(
            sindecmu_binning.name,
            sindecmu_binning.lower_edge,
            sindecmu_binning.upper_edge,
        ))

        self._hist_logE_sinDec = np.copy(pdf_log10emu_sindecmu)
        self.add_binning(log10emu_binning, name='log_energy')
        self.add_binning(sindecmu_binning, name='sin_dec')

    def assert_is_valid_for_trial_data(self, tdm):
        """Checks if this PDF covers the entire value range of the trail
        data events.

        Parameters
        ----------
        tdm : TrialDataManager instance
            The TrialDataManager instance holding the data events.
            The following data fields need to exist:

                'sin_dec'

                'log_energy'

        Raises
        ------
        ValueError
            If parts of the trial data is outside the value range of this
            PDF.
        """
        sindecmu = tdm.get_data('sin_dec')
        if np.min(sindecmu) < self.get_axis(0).vmin:
            raise ValueError(
                'The minimum sindecmu value %e of the trial data is lower '
                'than the minimum value of the PDF %e!' % (
                    np.min(sindecmu), self.get_axis(0).vmin))
        if np.max(sindecmu) > self.get_axis(0).vmax:
            raise ValueError(
                'The maximum sindecmu value %e of the trial data is larger '
                'than the maximum value of the PDF %e!' % (
                    np.max(sindecmu), self.get_axis(0).vmax))

        log10emu = tdm.get_data('log_energy')
        if np.min(log10emu) < self.get_axis(1).vmin:
            raise ValueError(
                'The minimum log10emu value %e of the trial data is lower '
                'than the minimum value of the PDF %e!' % (
                    np.min(log10emu), self.get_axis(1).vmin))
        if np.max(log10emu) > self.get_axis(1).vmax:
            raise ValueError(
                'The maximum log10emu value %e of the trial data is larger '
                'than the maximum value of the PDF %e!' % (
                    np.max(log10emu), self.get_axis(1).vmax))

    def get_prob(self, tdm, params=None, tl=None):
        """Gets the probability density for the given trial data events.

        Parameters
        ----------
        tdm : TrialDataManager instance
            The TrialDataManager instance holding the data events.
            The following data fields need to exist:

                'sin_dec'

                'log_energy'

        params : dict | None
            The dictionary containing the parameter names and values for which
            the probability should get calculated.
            By definition of this PDF, this is ``Ǹone``, because this PDF does
            not depend on any parameters.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        prob : (N_events,)-shaped numpy ndarray
            The 1D numpy ndarray with the probability density for each event.
        """
        get_data = tdm.get_data

        log10emu = get_data('log_energy')
        sindecmu = get_data('sin_dec')

        log10emu_idxs = np.digitize(
            log10emu, self.get_binning('log_energy').binedges) - 1
        sindecmu_idxs = np.digitize(
            sindecmu, self.get_binning('sin_dec').binedges) - 1

        with TaskTimer(tl, 'Evaluating sindecmu-log10emu PDF.'):
            pd = self._hist_logE_sinDec[(log10emu_idxs, sindecmu_idxs)]

        return pd



# -*- coding: utf-8 -*-

import numpy as np
from scipy import integrate

from skyllh.core.py import module_classname
from skyllh.core.debugging import get_logger
from skyllh.core.timing import TaskTimer
from skyllh.core.binning import get_bincenters_from_binedges
from skyllh.core.pdf import (
    PDF,
    PDFAxis,
    PDFSet,
    IsSignalPDF,
)
from skyllh.core.multiproc import (
    IsParallelizable,
    parallelize
)
from skyllh.core.parameters import (
    ParameterGrid,
    ParameterGridSet
)
from skyllh.i3.dataset import I3Dataset
from skyllh.physics.flux import FluxModel

from skyllh.analyses.i3.publicdata_ps.aeff import PDAeff
from skyllh.analyses.i3.publicdata_ps.utils import (
    FctSpline1D,
)
from skyllh.analyses.i3.publicdata_ps.smearing_matrix import (
    PDSmearingMatrix
)


class PDSignalEnergyPDF(PDF, IsSignalPDF):
    """This class provides a signal energy PDF for a spectrial index value.
    """
    def __init__(
            self, f_e_spl, **kwargs):
        """Creates a new signal energy PDF instance for a particular spectral
        index value.

        Parameters
        ----------
        f_e_spl : FctSpline1D instance
            The FctSpline1D instance representing the spline of the energy PDF.
        """
        super().__init__(**kwargs)

        if not isinstance(f_e_spl, FctSpline1D):
            raise TypeError(
                'The f_e_spl argument must be an instance of FctSpline1D!')

        self.f_e_spl = f_e_spl

        self.log10_reco_e_lower_binedges = self.f_e_spl.x_binedges[:-1]
        self.log10_reco_e_upper_binedges = self.f_e_spl.x_binedges[1:]

        self.log10_reco_e_min = self.log10_reco_e_lower_binedges[0]
        self.log10_reco_e_max = self.log10_reco_e_upper_binedges[-1]

        # Add the PDF axes.
        self.add_axis(PDFAxis(
            name='log_energy',
            vmin=self.log10_reco_e_min,
            vmax=self.log10_reco_e_max)
        )

        # Check integrity.
        integral = integrate.quad(
            self.f_e_spl.evaluate,
            self.log10_reco_e_min,
            self.log10_reco_e_max,
            limit=200,
            full_output=1
        )[0] / self.f_e_spl.norm
        if not np.isclose(integral, 1):
            raise ValueError(
                'The integral over log10_reco_e of the energy term must be '
                'unity! But it is {}!'.format(integral))

    def assert_is_valid_for_trial_data(self, tdm):
        pass

    def get_pd_by_log10_reco_e(self, log10_reco_e, tl=None):
        """Calculates the probability density for the given log10(E_reco/GeV)
        values using the spline representation of the PDF.

        Parameters
        ----------
        log10_reco_e : (n_log10_reco_e,)-shaped 1D numpy ndarray
            The numpy ndarray holding the log10(E_reco/GeV) values for which
            the energy PDF should get evaluated.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        pd : (N_events,)-shaped numpy ndarray
            The 1D numpy ndarray with the probability density for each event.
        """
        # Select events that actually have a signal energy PDF.
        # All other events will get zero signal probability density.
        m = (
            (log10_reco_e >= self.log10_reco_e_min) &
            (log10_reco_e < self.log10_reco_e_max)
        )

        with TaskTimer(tl, 'Evaluate PDSignalEnergyPDF'):
            pd = np.zeros((len(log10_reco_e),), dtype=np.double)
            pd[m] = self.f_e_spl(log10_reco_e[m]) / self.f_e_spl.norm

        return pd

    def get_prob(self, tdm, params=None, tl=None):
        """Calculates the probability density for the events given by the
        TrialDataManager.

        Parameters
        ----------
        tdm : TrialDataManager instance
            The TrialDataManager instance holding the data events for which the
            probability should be looked up. The following data fields are
            required:
                - 'log_energy'
                    The log10 of the reconstructed energy.
        params : dict | None
            The dictionary containing the parameter names and values for which
            the probability should get calculated.
            By definition this PDF does not depend on parameters.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        pd : (N_events,)-shaped numpy ndarray
            The 1D numpy ndarray with the probability density for each event.
        grads : (N_fitparams,N_events)-shaped ndarray | None
            The 2D numpy ndarray holding the gradients of the PDF w.r.t.
            each fit parameter for each event. The order of the gradients
            is the same as the order of floating parameters specified through
            the ``param_set`` property.
            It is ``None``, if this PDF does not depend on any parameters.
        """
        log10_reco_e = tdm.get_data('log_energy')

        pd = self.get_pd_by_log10_reco_e(log10_reco_e, tl=tl)

        return (pd, None)


class PDSignalEnergyPDFSet(PDFSet, IsSignalPDF, IsParallelizable):
    """This class provides a signal energy PDF set for the public data.
    It creates a set of PDSignalEnergyPDF instances, one for each spectral
    index value on a grid.
    """
    def __init__(
            self,
            ds,
            src_dec,
            flux_model,
            fitparam_grid_set,
            ncpu=None,
            ppbar=None,
            **kwargs):
        """Creates a new PDSignalEnergyPDFSet instance for the public data.

        Parameters
        ----------
        ds : I3Dataset instance
            The I3Dataset instance that defines the dataset of the public data.
        src_dec : float
            The declination of the source in radians.
        flux_model : FluxModel instance
            The FluxModel instance that defines the source's flux model.
        fitparam_grid_set : ParameterGrid | ParameterGridSet instance
            The parameter grid set defining the grids of the fit parameters.
        ncpu : int | None
            The number of CPUs to utilize. Global setting will take place if
            not specified, i.e. set to None.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar for the optional parent progress bar.
        """
        self._logger = get_logger(module_classname(self))

        # Check for the correct types of the arguments.
        if not isinstance(ds, I3Dataset):
            raise TypeError(
                'The ds argument must be an instance of I3Dataset!')

        if not isinstance(flux_model, FluxModel):
            raise TypeError(
                'The flux_model argument must be an instance of FluxModel!')

        if (not isinstance(fitparam_grid_set, ParameterGrid)) and\
           (not isinstance(fitparam_grid_set, ParameterGridSet)):
            raise TypeError(
                'The fitparam_grid_set argument must be an instance of type '
                'ParameterGrid or ParameterGridSet!')

        # Extend the fitparam_grid_set to allow for parameter interpolation
        # values at the grid edges.
        fitparam_grid_set = fitparam_grid_set.copy()
        fitparam_grid_set.add_extra_lower_and_upper_bin()

        super().__init__(
            pdf_type=PDF,
            fitparams_grid_set=fitparam_grid_set,
            ncpu=ncpu
        )

        # Load the smearing matrix.
        sm = PDSmearingMatrix(
            pathfilenames=ds.get_abs_pathfilename_list(
                ds.get_aux_data_definition('smearing_datafile')))

        # Select the slice of the smearing matrix corresponding to the
        # source declination band.
        true_dec_idx = sm.get_true_dec_idx(src_dec)
        sm_pdf = sm.pdf[:, true_dec_idx]

        # Only look at true neutrino energies for which a recostructed
        # muon energy distribution exists in the smearing matrix.
        (min_log_true_e,
         max_log_true_e) = sm.get_true_log_e_range_with_valid_log_e_pdfs(
            true_dec_idx)
        log_true_e_mask = np.logical_and(
            sm.log10_true_enu_binedges >= min_log_true_e,
            sm.log10_true_enu_binedges <= max_log_true_e)
        true_enu_binedges = np.power(
            10, sm.log10_true_enu_binedges[log_true_e_mask])
        true_enu_binedges_lower = true_enu_binedges[:-1]
        true_enu_binedges_upper = true_enu_binedges[1:]
        valid_true_e_idxs = [sm.get_log10_true_e_idx(0.5 * (he + le))
            for he,le in zip(
                sm.log10_true_enu_binedges[log_true_e_mask][1:],
                sm.log10_true_enu_binedges[log_true_e_mask][:-1])
            ]

        xvals_binedges = ds.get_binning_definition('log_energy').binedges
        xvals = get_bincenters_from_binedges(xvals_binedges)

        # Calculate the neutrino enegry bin widths in GeV.
        d_enu = np.diff(true_enu_binedges)
        self._logger.debug(
            'dE_nu = {}'.format(d_enu)
        )

        # Load the effective area.
        aeff = PDAeff(
            pathfilenames=ds.get_abs_pathfilename_list(
                ds.get_aux_data_definition('eff_area_datafile')))

        # Calculate the probability to detect a neutrino of energy
        # E_nu given a neutrino declination: p(E_nu|dec).
        det_prob = aeff.get_detection_prob_for_decnu(
            decnu=src_dec,
            enu_min=true_enu_binedges[:-1],
            enu_max=true_enu_binedges[1:],
            enu_range_min=true_enu_binedges[0],
            enu_range_max=true_enu_binedges[-1]
        )

        self._logger.debug('det_prob = {}, sum = {}'.format(
            det_prob, np.sum(det_prob)))

        if not np.isclose(np.sum(det_prob), 1):
            self._logger.warn(
                'The sum of the detection probabilities is not unity! It is '
                '{}.'.format(np.sum(det_prob)))

        psi_edges_bw = sm.psi_upper_edges - sm.psi_lower_edges
        ang_err_bw = sm.ang_err_upper_edges - sm.ang_err_lower_edges

        # Create the energy pdf for different gamma values.
        def create_energy_pdf(sm_pdf, flux_model, gridfitparams):
            """Creates an energy pdf for a specific gamma value.
            """
            # Create a copy of the FluxModel with the given flux parameters.
            # The copy is needed to not interfer with other CPU processes.
            my_flux_model = flux_model.copy(newprop=gridfitparams)

            self._logger.debug(
                'Generate signal energy PDF for parameters {} in {} E_nu '
                'bins.'.format(
                    gridfitparams, len(valid_true_e_idxs))
            )

            # Calculate the flux probability p(E_nu|gamma).
            flux_prob = (
                my_flux_model.get_integral(
                    true_enu_binedges_lower,
                    true_enu_binedges_upper
                ) /
                my_flux_model.get_integral(
                    true_enu_binedges[0],
                    true_enu_binedges[-1]
                )
            )
            if not np.isclose(np.sum(flux_prob), 1):
                self._logger.warn(
                    'The sum of the flux probabilities is not unity! It is '
                    '{}.'.format(np.sum(flux_prob)))

            self._logger.debug(
                'flux_prob = {}, sum = {}'.format(
                    flux_prob, np.sum(flux_prob))
            )

            p = flux_prob * det_prob

            true_e_prob = p / np.sum(p)

            self._logger.debug(
                'true_e_prob = {}'.format(
                    true_e_prob))

            def create_reco_e_pdf_for_true_e(idx, true_e_idx):
                """This functions creates a spline for the reco energy
                distribution given a true neutrino engery.
                """
                # Create the energy PDF f_e = P(log10_E_reco|dec) =
                # \int dPsi dang_err P(E_reco,Psi,ang_err).
                f_e = np.sum(
                    sm_pdf[true_e_idx] *
                    psi_edges_bw[true_e_idx, true_dec_idx, :, :, np.newaxis] *
                    ang_err_bw[true_e_idx, true_dec_idx, :, :, :],
                    axis=(-1, -2)
                )

                # Build the spline for this P(E_reco|E_nu). Weigh the pdf
                # with the true neutrino energy probability (flux prob).
                log10_reco_e_binedges = sm.log10_reco_e_binedges[
                    true_e_idx, true_dec_idx]

                p = f_e * true_e_prob[idx]

                spline = FctSpline1D(p, log10_reco_e_binedges)

                return spline(xvals)

            # Integrate over the true neutrino energy and spline the output.
            sum_pdf = np.sum([
                create_reco_e_pdf_for_true_e(i, true_e_idx)
                for i,true_e_idx in enumerate(valid_true_e_idxs)
            ], axis=0)

            spline = FctSpline1D(sum_pdf, xvals_binedges, norm=True)

            pdf = PDSignalEnergyPDF(spline)

            return pdf

        args_list = [
            ((sm_pdf, flux_model, gridfitparams), {})
            for gridfitparams in self.gridfitparams_list
        ]

        pdf_list = parallelize(
            create_energy_pdf,
            args_list,
            ncpu=self.ncpu,
            ppbar=ppbar)

        del(sm_pdf)

        # Save all the energy PDF objects in the PDFSet PDF registry with
        # the hash of the individual parameters as key.
        for (gridfitparams, pdf) in zip(self.gridfitparams_list, pdf_list):
            self.add_pdf(pdf, gridfitparams)

    def get_prob(self, tdm, gridfitparams, tl=None):
        """Calculates the signal probability density of each event for the
        given set of signal fit parameters on a grid.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the data events for which the
            probability should be calculated for. The following data fields must
            exist:

            - 'log_energy'
                The log10 of the reconstructed energy.
            - 'psi'
                The opening angle from the source to the event in radians.
            - 'ang_err'
                The angular error of the event in radians.
        gridfitparams : dict
            The dictionary holding the signal parameter values for which the
            signal energy probability should be calculated. Note, that the
            parameter values must match a set of parameter grid values for which
            a PDSignalPDF object has been created at construction time of this
            PDSignalPDFSet object.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure time.

        Returns
        -------
        prob : 1d ndarray
            The array with the signal energy probability for each event.
        grads : (N_fitparams,N_events)-shaped ndarray | None
            The 2D numpy ndarray holding the gradients of the PDF w.r.t.
            each fit parameter for each event. The order of the gradients
            is the same as the order of floating parameters specified through
            the ``param_set`` property.
            It is ``None``, if this PDF does not depend on any parameters.

        Raises
        ------
        KeyError
            If no energy PDF can be found for the given signal parameter values.
        """
        pdf = self.get_pdf(gridfitparams)

        (prob, grads) = pdf.get_prob(tdm, tl=tl)

        return (prob, grads)



# -*- coding: utf-8 -*-

import abc
import numpy as np

from skyllh.core.optimize import (
    AllEventSelectionMethod,
    EventSelectionMethod
)
from skyllh.core.py import (
    float_cast,
    func_has_n_args,
    issequenceof
)
from skyllh.core.debugging import get_logger
from skyllh.core.scrambling import DataScrambler
from skyllh.core.timing import TaskTimer
from skyllh.core.config import CFG


logger = get_logger(__name__)


class BackgroundGenerationMethod(object, metaclass=abc.ABCMeta):
    """This is the abstract base class for a detector specific background
    generation method.
    """

    def __init__(self):
        """Constructs a new background generation method instance.
        """
        super(BackgroundGenerationMethod, self).__init__()

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Notifies the background generation method about an updated
        SourceHypoGroupManager instance.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The new SourceHypoGroupManager instance.
        """
        pass

    @abc.abstractmethod
    def generate_events(self, rss, dataset, data, mean, tl=None, **kwargs):
        """This method is supposed to generate a `mean` number of background
        events for the given dataset and its data.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        dataset : instance of Dataset
            The Dataset instance describing the dataset for which background
            events should get generated.
        data : instance of DatasetData
            The DatasetData instance holding the data of the dataset for which
            background events should get generated.
        mean : float
            The mean number of background events to generate.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.
        **kwargs
            Additional keyword arguments, which might be required for a
            particular background generation method.

        Returns
        -------
        n_bkg : int
            The number of generated background events.
        bkg_events : DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events. The number of events in this array might be less
            than `n_bkg` if an event selection method was used for optimization
            purposes. The difference `n_bkg - len(bkg_events)` is then the
            number of pure background events in the generated background event
            sample.
        """
        pass


class MCDataSamplingBkgGenMethod(BackgroundGenerationMethod):
    """This class implements the method to generate background events from
    monte-carlo (MC) data by sampling events from the MC data set according to a
    probability value given for each event. Functions can be provided to get the
    mean number of background events and the probability of each monte-carlo
    event.
    """
    def __init__(
        self, get_event_prob_func, get_mean_func=None, unique_events=False,
        data_scrambler=None, mc_inplace_scrambling=False,
        keep_mc_data_fields=None, pre_event_selection_method=None):
        """Creates a new instance of the MCDataSamplingBkgGenMethod class.

        Parameters
        ----------
        get_event_prob_func : callable
            The function to get the background probability of each monte-carlo
            event. The call signature of this function must be
            `__call__(dataset, data, events)`,
            where `dataset` and `data` are `Dataset` and `DatasetData` instances
            of the data set for which background events needs to get generated.
            The `events` argument holds the actual set of events, for which the
            background event probabilities need to get calculated.
        get_mean_func : callable | None
            The function to get the mean number of background events.
            The call signature of this function must be
            `__call__(dataset, data, events)`,
            where `dataset` and `data` are `Dataset` and `DatasetData` instances
            of the data set for which background events needs to get generated.
            The `events` argument holds the actual set of events, for which the
            mean number of background events should get calculated.
            This argument can be `None`, which means that the mean number of
            background events to generate needs to get specified through the
            `generate_events` method. However, if an event selection method is
            provided, this argument cannot be None.
        unique_events : bool
            Flag if unique events should be drawn from the monte-carlo (True),
            or if events can be drawn several times (False). Default is False.
        data_scrambler : instance of DataScrambler | None
            If set to an instance of DataScrambler, the drawn monte-carlo
            background events will get scrambled. This can ensure more
            independent data trials. It is especially important when monte-carlo
            statistics are low.
        mc_inplace_scrambling : bool
            Flag if the scrambling of the monte-carlo data should be done
            inplace, i.e. without creating a copy of the MC data first.
            Default is False.
        keep_mc_data_fields : str | list of str | None
            The MC data field names that should be kept in order to be able to
            calculate the background events rates by the functions
            ``get_event_prob_func`` and ``get_mean_func``. All other MC fields
            will get droped due to computational efficiency reasons.
        pre_event_selection_method : instance of EventSelectionMethod | None
            If set to an instance of EventSelectionMethod, this method will
            pre-select the MC events that will be used for later background
            event generation. Using this pre-selection a large portion of the
            MC data can be reduced prior to background event generation.
        """
        super(MCDataSamplingBkgGenMethod, self).__init__()

        self.get_event_prob_func = get_event_prob_func
        self.get_mean_func = get_mean_func
        self.unique_events = unique_events
        self.data_scrambler = data_scrambler
        self.mc_inplace_scrambling = mc_inplace_scrambling
        self.keep_mc_data_field_names = keep_mc_data_fields
        self.pre_event_selection_method = pre_event_selection_method

        if((pre_event_selection_method is not None) and (get_mean_func is None)):
            raise ValueError('If an event pre-selection method is provided, a '
                             'get_mean_func needs to be provided as well!')

        # Define cache members to cache the background probabilities for each
        # monte-carlo event. The probabilities change only if the data changes.
        self._cache_data_id = None
        self._cache_mc_pre_selected = None
        self._cache_mc_event_bkg_prob = None
        self._cache_mc_event_bkg_prob_pre_selected = None
        self._cache_mean = None

    @property
    def get_event_prob_func(self):
        """The function to obtain the background probability for each
        monte-carlo event of the data set.
        """
        return self._get_event_prob_func
    @get_event_prob_func.setter
    def get_event_prob_func(self, func):
        if(not callable(func)):
            raise TypeError('The get_event_prob_func property must be a '
                'callable!')
        if(not func_has_n_args(func, 3)):
            raise TypeError('The function provided for the get_event_prob_func '
                'property must have 3 arguments!')
        self._get_event_prob_func = func

    @property
    def get_mean_func(self):
        """The function to obtain the mean number of background events for the
        data set. This can be None, which means that the mean number of
        background events to generate needs to be specified through the
        `generate_events` method.
        """
        return self._get_mean_func
    @get_mean_func.setter
    def get_mean_func(self, func):
        if(func is not None):
            if(not callable(func)):
                raise TypeError('The get_mean_func property must be a '
                    'callable!')
            if(not func_has_n_args(func, 3)):
                raise TypeError('The function provided for the get_mean_func '
                    'property must have 3 arguments!')
        self._get_mean_func = func

    @property
    def unique_events(self):
        """Flag if unique events should be drawn from the monto-carlo (True),
        or if the same event can be drawn multiple times from the monte-carlo.
        """
        return self._unique_events
    @unique_events.setter
    def unique_events(self, b):
        if(not isinstance(b, bool)):
            raise TypeError('The unique_events property must be of type bool!')
        self._unique_events = b

    @property
    def data_scrambler(self):
        """The DataScrambler instance that should be used to scramble the drawn
        monte-carlo background events to ensure more independent data trials.
        This is especially important when monte-carlo statistics are low. It is
        `None`, if no data scrambling should be used.
        """
        return self._data_scrambler
    @data_scrambler.setter
    def data_scrambler(self, scrambler):
        if(scrambler is not None):
            if(not isinstance(scrambler, DataScrambler)):
                raise TypeError('The data_scrambler property must be an instance '
                    'of DataScrambler!')
        self._data_scrambler = scrambler

    @property
    def mc_inplace_scrambling(self):
        """Flag if the scrambling of the monte-carlo data should be done
        inplace, i.e. without creating a copy of the MC data first.
        """
        return self._mc_inplace_scrambling
    @mc_inplace_scrambling.setter
    def mc_inplace_scrambling(self, b):
        if(not isinstance(b, bool)):
            raise TypeError('The mc_inplace_scrambling property must be of '
                'type bool!')
        self._mc_inplace_scrambling = b

    @property
    def keep_mc_data_field_names(self):
        """The MC data field names that should be kept in order to be able to
        calculate the background events rates by the functions
        ``get_event_prob_func`` and ``get_mean_func``. All other MC fields
        will get droped due to computational efficiency reasons.
        """
        return self._keep_mc_data_field_names
    @keep_mc_data_field_names.setter
    def keep_mc_data_field_names(self, names):
        if(names is None):
            names = []
        elif(isinstance(names, str)):
            names = [ names ]
        elif(not issequenceof(names, str)):
            raise TypeError('The keep_mc_data_field_names must be None, an '
                'instance of type str, or a sequence of objects of type str!')
        self._keep_mc_data_field_names = names

    @property
    def pre_event_selection_method(self):
        """The instance of EventSelectionMethod that pre-selects the MC events,
        which can be considered for background event generation.
        """
        return self._pre_event_selection_method
    @pre_event_selection_method.setter
    def pre_event_selection_method(self, method):
        if(method is not None):
            if(not isinstance(method, EventSelectionMethod)):
                raise TypeError('The pre_event_selection_method property must '
                    'be None, or an instance of EventSelectionMethod!')
            # If the event selection method selects all events, it's equivalent
            # to have it set to None, because then no operation has to be
            # performed.
            if(isinstance(method, AllEventSelectionMethod)):
                method = None
        self._pre_event_selection_method = method

    def change_source_hypo_group_manager(self, src_hypo_group_manager):
        """Changes the SourceHypoGroupManager instance of the
        pre-event-selection method. Also it invalides the data cache of this
        background generation method.

        Parameters
        ----------
        src_hypo_group_manager : SourceHypoGroupManager instance
            The new SourceHypoGroupManager instance.
        """
        if(self._pre_event_selection_method is not None):
            self._pre_event_selection_method.change_source_hypo_group_manager(
                src_hypo_group_manager)

        # Invalidate the data cache.
        self._cache_data_id = None

    def generate_events(
            self, rss, dataset, data, mean=None, poisson=True, tl=None):
        """Generates a `mean` number of background events for the given dataset
        and its data.

        Parameters
        ----------
        rss : instance of RandomStateService
            The instance of RandomStateService that should be used to generate
            random numbers from.
        dataset : instance of Dataset
            The Dataset instance describing the dataset for which background
            events should get generated.
        data : instance of DatasetData
            The DatasetData instance holding the data of the dataset for which
            background events should get generated.
        mean : float | None
            The mean number of background events to generate.
            Can be `None`. In that case the mean number of background events is
            obtained through the `get_mean_func` function.
        poisson : bool
            If set to True (default), the actual number of generated background
            events will be drawn from a Poisson distribution with the given mean
            value of background events.
            If set to False, the argument ``mean`` specifies the actual number
            of generated background events.
        tl : instance of TimeLord | None
            The optional instance of TimeLord that should be used to collect
            timing information about this method.

        Returns
        -------
        n_bkg : int
            The number of generated background events for the data set.
        bkg_events : instance of DataFieldRecordArray
            The instance of DataFieldRecordArray holding the generated
            background events. The number of events can be less than `n_bkg`
            if an event selection method is used.
        """
        tracing_enabled = CFG['debugging']['enable_tracing']

        # Create aliases to avoid dot-lookup.
        self__pre_event_selection_method = self._pre_event_selection_method

        # Check if the data set has changed. In that case need to get new
        # background probabilities for each monte-carlo event and a new mean
        # number of background events.
        data_id = id(data)
        if(self._cache_data_id != data_id):
            if(tracing_enabled):
                logger.debug(
                    f'DatasetData instance id of dataset "{dataset.name}" '
                    f'changed from {self._cache_data_id} to {data_id}')
            # Cache the current id of the data.
            self._cache_data_id = data_id

            # Create a copy of the MC data with all MC data fields removed,
            # except the specified MC data fields to keep for the
            # ``get_mean_func`` and ``get_event_prob_func`` functions.
            keep_field_names = list(set(
                CFG['dataset']['analysis_required_exp_field_names'] +
                data.exp_field_names +
                self._keep_mc_data_field_names
            ))
            data_mc = data.mc.copy(keep_fields=keep_field_names)

            if(self._get_mean_func is not None):
                with TaskTimer(tl, 'Calculate total MC background mean.'):
                    self._cache_mean = self._get_mean_func(
                        dataset, data, data_mc)

            with TaskTimer(tl, 'Calculate MC background event probability cache.'):
                self._cache_mc_event_bkg_prob = self._get_event_prob_func(
                    dataset, data, data_mc)

            if(self__pre_event_selection_method is not None):
                with TaskTimer(tl, 'Pre-select MC events.'):
                    (self._cache_mc_pre_selected,
                     mc_pre_selected_mask_idxs) =\
                    self__pre_event_selection_method.select_events(
                        data_mc, ret_mask_idxs=True, tl=tl)
                self._cache_mc_event_bkg_prob_pre_selected = self._cache_mc_event_bkg_prob[mc_pre_selected_mask_idxs]
            else:
                self._cache_mc_pre_selected = data_mc


        if(mean is None):
            if(self._cache_mean is None):
                raise ValueError('No mean number of background events and no '
                    'get_mean_func were specified! One of the two must be '
                    'specified!')
            mean = self._cache_mean
        else:
            mean = float_cast(mean, 'The mean number of background events must '
                'be castable to type float!')

        # Draw the number of background events from a poisson distribution with
        # the given mean number of background events. This will be the number of
        # background events for this data set.
        n_bkg = (int(rss.random.poisson(mean)) if poisson else
                 int(np.round(mean, 0)))

        # Apply only event pre-selection before choosing events.
        data_mc_selected = self._cache_mc_pre_selected

        # Calculate the mean number of background events for the pre-selected
        # MC events.
        if(self__pre_event_selection_method is None):
            # No selection at all, use the total mean.
            mean_selected = mean
        else:
            with TaskTimer(tl, 'Calculate selected MC background mean.'):
                mean_selected = self._get_mean_func(
                    dataset, data, data_mc_selected)

        # Calculate the actual number of background events for the selected
        # events.
        p_binomial = mean_selected / mean
        with TaskTimer(tl, 'Get p array.'):
            if(self__pre_event_selection_method is None):
                p = self._cache_mc_event_bkg_prob
            else:
                # Pre-selection.
                p = self._cache_mc_event_bkg_prob_pre_selected / p_binomial
        n_bkg_selected = int(np.around(n_bkg * p_binomial, 0))

        # Draw the actual background events from the selected events of the
        # monto-carlo data set.
        with TaskTimer(tl, 'Draw MC background indices.'):
            bkg_event_indices = rss.random.choice(
                data_mc_selected.indices,
                size=n_bkg_selected,
                p=p,
                replace=(not self._unique_events))
        with TaskTimer(tl, 'Select MC background events from indices.'):
            bkg_events = data_mc_selected[bkg_event_indices]

        # Scramble the drawn MC events if requested.
        if(self._data_scrambler is not None):
            with TaskTimer(tl, 'Scramble MC background data.'):
                bkg_events = self._data_scrambler.scramble_data(
                    rss, bkg_events, copy=False)

        # Remove MC specific data fields from the background events record
        # array. So the result contains only experimental data fields. The list
        # of experimental data fields is defined as the unique set of the
        # required experimental data fields defined by the data set, and the
        # actual experimental data fields (in case there are additional kept
        # data fields by the user).
        with TaskTimer(tl, 'Remove MC specific data fields from MC events.'):
            exp_field_names = list(set(
                CFG['dataset']['analysis_required_exp_field_names'] +
                data.exp_field_names))
            bkg_events.tidy_up(exp_field_names)

        return (n_bkg, bkg_events)



# -*- coding: utf-8 -*-

from __future__ import division

import abc
import numpy as np

from skyllh.core.py import float_cast, classname

class TimeProfileModel(object, metaclass=abc.ABCMeta):
    """Abstract base class for an emission time profile of a source.
    """

    def __init__(self, t_start, t_end):
        """Creates a new time profile instance.

        Parameters
        ----------
        t_start : float
            The MJD start time of the box profile.
        t_end : float
            The MJD end time of the box profile.
        """
        super(TimeProfileModel, self).__init__()

        self.t_start = t_start
        self.t_end = t_end

    @property
    def t_start(self):
        """The MJD start time of the box profile.
        """
        return self._t_start
    @t_start.setter
    def t_start(self, t):
        t = float_cast(t,
            'The t_start property must be castable to type float!'
        )
        self._t_start = t

    @property
    def t_end(self):
        """The MJD end time of the box profile.
        """
        return self._t_end
    @t_end.setter
    def t_end(self, t):
        t = float_cast(t,
            'The t_end property must be castable to type float!'
        )
        self._t_end = t

    @property
    def duration(self):
        """The duration (in days) of the time profile.
        """
        return self._t_end - self._t_start

    @abc.abstractmethod
    def move(self, dt):
        """Abstract method to move the time profile by the given amount of time.

        Parameters
        ----------
        dt : float
            The MJD time difference of how far to move the time profile in time.
            This can be a positive or negative time shift.
        """
        pass

    @abc.abstractmethod
    def update(self, fitparams):
        """This method is supposed to update the time profile based on new
        fit parameter values. This method should return a boolean flag if an
        update was actually performed.

        Returns
        -------
        updated : bool
            Flag if the time profile actually got updated because of new fit
            parameter values.
        """
        pass

    @abc.abstractmethod
    def get_integral(self, t1, t2):
        """This method is supposed to calculate the integral of the time profile
        from time t1 to time t2.

        Parameters
        ----------
        t1 : float | array of float
            The MJD start time of the integration.
        t2 : float | array of float
            The MJD end time of the integration.

        Returns
        -------
        integral : array of float
            The integral value(s) of the time profile.
        """
        pass

    @abc.abstractmethod
    def get_total_integral(self):
        """This method is supposed to calculate the total integral of the time
        profile from t_start to t_end.

        Returns
        -------
        integral : float
            The integral value of the entire time profile.
        """
        pass

    @abc.abstractmethod
    def get_value(self, t):
        """Retrieves the value of the time profile at time t.

        Parameters
        ----------
        t : float
            The MJD time for which the time profile value should get retrieved.

        Returns
        -------
        value : float
            The time profile value at the given time.
        """
        pass


class BoxTimeProfile(TimeProfileModel):
    """The BoxTimeProfile describes a box-shaped emission time profile of a
    source. It has the following fit parameters:

        T0 : float
            The mid MJD time of the box profile.
        Tw : float
            The width (days) of the box profile.
    """
    def __init__(self, T0, Tw):
        """Creates a new box-shaped time profile instance.

        Parameters
        ----------
        T0 : float
            The mid MJD time of the box profile.
        Tw : float
            The width (days) of the box profile.
        """
        t_start = T0 - Tw/2.
        t_end = T0 + Tw/2.

        super(BoxTimeProfile, self).__init__(t_start, t_end)

    def move(self, dt):
        """Moves the box-shaped time profile by the time difference dt.

        Parameters
        ----------
        dt : float
            The MJD time difference of how far to move the time profile in time.
            This can be a positive or negative time shift.
        """
        self._t_start += dt
        self._t_end += dt

    @property
    def T0(self):
        """The time of the mid point of the box.
        """
        return 0.5*(self._t_start + self._t_end)
    @T0.setter
    def T0(self, t):
        old_T0 = self.T0
        dt = t - old_T0
        self.move(dt)

    @property
    def Tw(self):
        """The time width (in days) of the box.
        """
        return self._t_end - self._t_start
    @Tw.setter
    def Tw(self, w):
        T0 = self.T0
        self._t_start = T0 - 0.5*w
        self._t_end = T0 + 0.5*w

    def __str__(self):
        """Pretty string representation of the BoxTimeProfile class instance.
        """
        s = '%s(T0=%.6f, Tw=%.6f)'%(classname(self), self.T0, self.Tw)
        return s

    def update(self, fitparams):
        """Updates the box-shaped time profile with the new fit parameter
        values.

        Parameters
        ----------
        fitparams : dict
            The dictionary with the new fit parameter values. The key must be
            the name of the fit parameter and the value the new parameter value.

        Returns
        -------
        updated : bool
            Flag if the time profile actually got updated because of new fit
            parameter values.
        """
        updated = False

        self_T0 = self.T0
        T0 = fitparams.get('T0', self_T0)
        if(T0 != self_T0):
            self.T0 = T0
            updated = True

        self_Tw = self.Tw
        Tw = fitparams.get('Tw', self_Tw)
        if(Tw != self_Tw):
            self.Tw = Tw
            updated = True

        return updated

    def get_integral(self, t1, t2):
        """Calculates the integral of the box-shaped time profile from MJD time
        t1 to time t2.

        Parameters
        ----------
        t1 : float | array of float
            The MJD start time(s) of the integration.
        t2 : float | array of float
            The MJD end time(s) of the integration.

        Returns
        -------
        integral : array of float
            The integral value(s).
        """
        t1 = np.atleast_1d(t1)
        t2 = np.atleast_1d(t2)

        integrals = np.zeros((t1.shape[0],), dtype=np.float64)

        m = (t2 > self._t_start) & (t1 < self._t_end)
        N = np.count_nonzero(m)

        t1 = np.max(np.vstack((t1[m], np.repeat(self._t_start, N))).T, axis=1)
        t2 = np.min(np.vstack((t2[m], np.repeat(self._t_end, N))).T, axis=1)

        f = 1./self.duration

        integrals[m] = f*(t2-t1)

        return integrals

    def get_total_integral(self):
        """Calculates the the total integral of the box-shaped time profile
        from t_start to t_end. By definition it is 1.

        Returns
        -------
        integral : float
            The integral value of the entire time profile.
        """
        return 1.

    def get_value(self, t):
        """Retrieves the value of the box-shaped time profile at time t.
        For a box-shaped time profile the values are all equal to 1/duration
        for times within the time duration and zero otherwise.

        Parameters
        ----------
        t : float | array of float
            The MJD time(s) for which the time profile value(s) should get
            retrieved.

        Returns
        -------
        values : array of float
            The time profile value(s) at the given time(s).
        """
        t = np.atleast_1d(t)

        values = np.zeros((t.shape[0],), dtype=np.float64)
        m = (t >= self._t_start) & (t < self._t_end)
        values[m] = 1./self.duration

        return values



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np
import sys
import time

from skyllh.core import display
from skyllh.core import session
from skyllh.core.py import int_cast


class ProgressBar(object):
    """This class implements a hierarchical progress bar that can serve as a
    parent for child progress bars to display different levels of processing.
    """
    def __init__(self, maxval, startval=0, parent=None):
        """Creates a new ProgressBar instance.

        Parameters
        ----------
        maxval : int
            The maximal value the progress can reach.
        startval : int
            The progress value to start with. Must be smaller than `maxval`.
        parent : instance of ProgressBar
            The parent instance of ProgressBar if this progress bar is a sub
            progress bar.
        """
        super(ProgressBar, self).__init__()

        self.maxval = maxval
        self.startval = startval
        self.parent = parent

        self._start_time = None
        self._val = 0
        self._last_rendered_pbar_str = None
        self._sub_pbar_list = []

    @property
    def maxval(self):
        """The maximal integer value the progress can reach.
        """
        return self._maxval
    @maxval.setter
    def maxval(self, v):
        v = int_cast(v, 'The maxval property must be castable to an integer '
            'value!')
        self._maxval = v

    @property
    def startval(self):
        """The progress integer value to start with. It must be smaller than
        `maxval`.
        """
        return self._startval
    @startval.setter
    def startval(self, v):
        v = int_cast(v, 'The startval property must be castable to an integer '
            ' value!')
        if(v >= self._maxval):
            raise ValueError('The startval value (%d) must be smaller than the '
                'value of the `maxval` property (%d)!',
                v, self._maxval)
        self._startval = v

    @property
    def parent(self):
        """The parent ProgressBar instance of this progress bar, or `None` if
        no parent exist.
        """
        return self._parent
    @parent.setter
    def parent(self, pbar):
        if(pbar is not None):
            if(not isinstance(pbar, ProgressBar)):
                raise TypeError('The parent property must be None, or an '
                    'instance of ProgressBar!')
        self._parent = pbar

    @property
    def progress(self):
        """(read-only) The progress of this progress bar as a number between 0
        and 1.
        """
        return (self._val - self._startval) / (self._maxval - self._startval)

    @property
    def gets_shown(self):
        """(read-only) Flag if the progress bar gets shown. This is ``True``
        if the program is run in an interactive session, ``False`` otherwise.
        """
        return session.is_interactive_session()

    def add_sub_progress_bar(self, pbar):
        """Adds the given progress bar to the list of running sub progress bars
        of this progress bar.
        """
        if(not isinstance(pbar, ProgressBar)):
            raise TypeError('The pbar argument must be an instance of '
                'ProgressBar!')
        self._sub_pbar_list.append(pbar)

    def remove_sub_progress_bar(self, pbar):
        """Removes the given progress bar instance from the list of running sub
        progress bars of this progress bar.
        """
        self._sub_pbar_list.remove(pbar)

    def _sec_to_hms(self, t):
        """Calculate hours, minutes, and seconds from `t` given in seconds.

        Parameters
        ----------
        t : float
            The time in seconds.

        Returns
        -------
        t_h : float
            The integer hours of `t`.
        t_m : float
            The integer minutes of `t`.
        t_s : float
            The integer seconds of `t`.
        """
        t_h = int(t / 3600)
        t -= t_h*3600
        t_m = int(t / 60)
        t -= t_m*60
        t_s = int(np.round(t, 0))

        return (t_h, t_m, t_s)

    def _render_pbar_str(self):
        """Renders the progress bar string. This method is called only when
        this progress bar has no parent.

        Returns
        -------
        pbar_str : str
            The rendered progress bar string.
        """
        # Calculate the elapsed time (ELT) for the first 10 seconds or if we
        # are at the end of the progress. Otherwise calculate the estimated
        # remaining time (ERT).
        curr_time = time.time()
        t_elapsed = curr_time - self._start_time

        t_label = 'ELT'
        t = t_elapsed

        progress = self.progress
        if(progress > 0 and (t_elapsed >= 10) and (self._val < self._maxval)):
            t_label = 'ERT'
            t_total = t_elapsed / progress
            t = t_total - t_elapsed

        (t_h, t_m, t_s) = self._sec_to_hms(t)

        # Get the current progress values from the sub progress bars and this
        # progress bar. The value of this progress bar is the last value.
        p_list = self.get_progress_list()

        sub_pbar_str = ''
        for p in p_list[:-1]:
            if(p != 1):
                sub_pbar_str += '%d '%(int(p*10))
            else:
                sub_pbar_str += '# '

        barwidth = max(display.PAGE_WIDTH - 22 - len(sub_pbar_str), 10)
        nchar = int(np.round(p_list[-1] * barwidth, 0))

        fmt = "%s[%-"+str(barwidth)+"s] %3d%% %s %dh:%02dm:%02ds"
        pbar_str = fmt%(
            sub_pbar_str,
            '='*nchar,
            int(np.round(p_list[-1]*100, 0)),
            t_label, t_h, t_m, t_s
        )

        return pbar_str

    def get_progress_list(self):
        """Retrieves the list of progress values of all the sub progress bars
        and this progress. The value of this progress bar is the last value.
        """
        p_list = []
        for pbar in self._sub_pbar_list:
            p_list.extend(pbar.get_progress_list())
        p_list.append(self.progress)

        return p_list

    def rerender(self):
        """Rerenders this progress bar on the display, but only if the rendered
        progress bar string changed.
        """
        pbar_str = self._render_pbar_str()
        if(pbar_str == self._last_rendered_pbar_str):
            return

        sys.stdout.write('\r'+pbar_str)
        sys.stdout.flush()
        self._last_rendered_pbar_str = pbar_str

    def trigger_rerendering(self):
        """Triggers a rerendering of the most top progress bar.
        """
        parent = self._parent
        if(parent is not None):
            parent.trigger_rerendering()
            return

        if(not session.is_interactive_session()):
            return

        # We are the most top parent progress bar. So we need to get rerendered.
        self.rerender()

    def start(self):
        """Starts the progress bar by taking the start time and setting the
        progress value to the start value.
        If this progress bar has a parent, it adds itself to the parent's list
        of running sub progress bars. Otherwise it will render and print this
        progress bar for the first time.

        Returns
        -------
        self : instance of ProgressBar
            The instance of this ProgressBar.
        """
        self._start_time = time.time()
        self._val = self._startval

        parent = self._parent
        if(parent is not None):
            # Add this progress bar to the parent progress bar.
            parent.add_sub_progress_bar(self)
            return self

        if(not session.is_interactive_session()):
            return self

        self._last_rendered_pbar_str = self._render_pbar_str()

        sys.stdout.write(self._last_rendered_pbar_str)
        sys.stdout.flush()

        return self

    def finish(self):
        """Finishes this progress bar by setting the current progress value to
        its max value.
        If this progress bar has a parent, it triggers a rerendering of the
        parent and then removes itself from the parent's list of running sub
        progress bars. Otherwise it will render and print this progress bar for
        the last time.
        """
        self._val = self._maxval

        parent = self._parent
        if(parent is not None):
            parent.trigger_rerendering()
            # Remove this progress bar from the parent progress bar.
            parent.remove_sub_progress_bar(self)
            return

        if(not session.is_interactive_session()):
            return

        self._last_rendered_pbar_str = self._render_pbar_str()

        sys.stdout.write('\r'+self._last_rendered_pbar_str+"\n")
        sys.stdout.flush()

    def update(self, val):
        """Updates the progress value. If this progress bar has no parent
        progress bar, a rerendering of this progress bar is made.

        Parameters
        ----------
        val : int
            The new current progress value.
        """
        self._val = val

        self.trigger_rerendering()

    def increment(self, dval=1):
        """Updates the progress bar by incrementing the progress by the given
        integral amount.

        Parameters
        ----------
        dval : int
            The amount of progress to increment the progress bar with.
        """
        self.update(self._val + dval)



# -*- coding: utf-8 -*-

import abc
import numpy as np
import scipy.interpolate

from skyllh.core.parameters import make_params_hash
from skyllh.core.multiproc import IsParallelizable, parallelize
from skyllh.core.pdfratio import SigSetOverBkgPDFRatio, PDFRatioFillMethod, MostSignalLikePDFRatioFillMethod

from skyllh.i3.pdf import I3EnergyPDF


class I3EnergySigSetOverBkgPDFRatioSpline(SigSetOverBkgPDFRatio, IsParallelizable):
    """This class implements a signal over background PDF ratio spline for
    I3EnergyPDF enegry PDFs. It takes an object, which is derived from PDFSet
    for I3EnergyPDF PDF types, and which is derived from IsSignalPDF, as signal
    PDF. Furthermore, it takes an object, which is derived from I3EnergyPDF and
    IsBackgroundPDF, as background PDF, and creates a spline for the ratio of
    the signal and background PDFs for a grid of different discrete energy
    signal fit parameters, which are defined by the signal PDF set.
    """
    def __init__(
            self, signalpdfset, backgroundpdf,
            fillmethod=None, interpolmethod=None, ncpu=None, ppbar=None):
        """Creates a new IceCube signal-over-background energy PDF ratio object.

        Parameters
        ----------
        signalpdfset : class instance derived from PDFSet (for PDF type
                       I3EnergyPDF), IsSignalPDF, and UsesBinning
            The PDF set, which provides signal energy PDFs for a set of
            discrete signal parameters.
        backgroundpdf : class instance derived from I3EnergyPDF, and
                        IsBackgroundPDF
            The background energy PDF object.
        fillmethod : instance of PDFRatioFillMethod | None
            An instance of class derived from PDFRatioFillMethod that implements
            the desired ratio fill method.
            If set to None (default), the default ratio fill method
            MostSignalLikePDFRatioFillMethod will be used.
        interpolmethod : class of GridManifoldInterpolationMethod
            The class implementing the fit parameter interpolation method for
            the PDF ratio manifold grid.
        ncpu : int | None
            The number of CPUs to use to create the ratio splines for the
            different sets of signal parameters.
        ppbar : ProgressBar instance | None
            The instance of ProgressBar of the optional parent progress bar.

        Raises
        ------
        ValueError
            If the signal and background PDFs use different binning.
        """
        super(I3EnergySigSetOverBkgPDFRatioSpline, self).__init__(
            pdf_type=I3EnergyPDF,
            signalpdfset=signalpdfset, backgroundpdf=backgroundpdf,
            interpolmethod=interpolmethod,
            ncpu=ncpu)

        # Define the default ratio fill method.
        if(fillmethod is None):
            fillmethod = MostSignalLikePDFRatioFillMethod()
        self.fillmethod = fillmethod

        # Ensure same binning of signal and background PDFs.
        for (sigpdf_hash, sigpdf) in self.signalpdfset.items():
            if(not sigpdf.has_same_binning_as(self.backgroundpdf)):
                raise ValueError('At least one signal PDF does not have the same binning as the background PDF!')

        def create_log_ratio_spline(sigpdfset, bkgpdf, fillmethod, gridfitparams):
            """Creates the signal/background ratio spline for the given signal
            parameters.

            Returns
            -------
            log_ratio_spline : RegularGridInterpolator
                The spline of the logarithmic PDF ratio values.
            """
            # Get the signal PDF for the given signal parameters.
            sigpdf = sigpdfset.get_pdf(gridfitparams)

            # Create the ratio array with the same shape than the background pdf
            # histogram.
            ratio = np.ones_like(bkgpdf.hist, dtype=np.float64)

            # Fill the ratio array.
            ratio = fillmethod.fill_ratios(ratio,
                sigpdf.hist, bkgpdf.hist,
                sigpdf.hist_mask_mc_covered, sigpdf.hist_mask_mc_covered_zero_physics,
                bkgpdf.hist_mask_mc_covered, bkgpdf.hist_mask_mc_covered_zero_physics)

            # Define the grid points for the spline. In general, we use the bin
            # centers of the binning, but for the first and last point of each
            # dimension we use the lower and upper bin edge, respectively, to
            # ensure full coverage of the spline across the binning range.
            points_list = []
            for binning in sigpdf.binnings:
                points = binning.bincenters
                (points[0], points[-1]) = (binning.lower_edge, binning.upper_edge)
                points_list.append(points)

            # Create the spline for the ratio values.
            log_ratio_spline = scipy.interpolate.RegularGridInterpolator(
                tuple(points_list),
                np.log(ratio),
                method='linear',
                bounds_error=False,
                fill_value=0.)

            return log_ratio_spline

        # Get the list of fit parameter permutations on the grid for which we
        # need to create PDF ratio arrays.
        gridfitparams_list = self.signalpdfset.gridfitparams_list

        args_list = [ ((signalpdfset, backgroundpdf, self.fillmethod, gridfitparams),{})
                     for gridfitparams in gridfitparams_list ]

        log_ratio_spline_list = parallelize(
            create_log_ratio_spline, args_list, self.ncpu, ppbar=ppbar)

        # Save all the log_ratio splines in a dictionary.
        self._gridfitparams_hash_log_ratio_spline_dict = dict()
        for (gridfitparams, log_ratio_spline) in zip(gridfitparams_list, log_ratio_spline_list):
            gridfitparams_hash = make_params_hash(gridfitparams)
            self._gridfitparams_hash_log_ratio_spline_dict[gridfitparams_hash] = log_ratio_spline

        # Save the list of data field names.
        self._data_field_names = [ binning.name
                                  for binning in self.backgroundpdf.binnings ]

        # Construct the instance for the fit parameter interpolation method.
        self._interpolmethod_instance = self.interpolmethod(self._get_spline_value, signalpdfset.fitparams_grid_set)

        # Create cache variables for the last ratio value and gradients in order
        # to avoid the recalculation of the ratio value when the
        # ``get_gradient`` method is called (usually after the ``get_ratio``
        # method was called).
        self._cache_fitparams_hash = None
        self._cache_ratio = None
        self._cache_gradients = None

    @property
    def fillmethod(self):
        """The PDFRatioFillMethod object, which should be used for filling the
        PDF ratio bins.
        """
        return self._fillmethod
    @fillmethod.setter
    def fillmethod(self, obj):
        if(not isinstance(obj, PDFRatioFillMethod)):
            raise TypeError('The fillmethod property must be an instance of PDFRatioFillMethod!')
        self._fillmethod = obj

    def _get_spline_value(self, tdm, gridfitparams, eventdata):
        """Selects the spline object for the given fit parameter grid point and
        evaluates the spline for all the given events.
        """
        # Get the spline object for the given fit parameter grid values.
        gridfitparams_hash = make_params_hash(gridfitparams)
        spline = self._gridfitparams_hash_log_ratio_spline_dict[gridfitparams_hash]

        # Evaluate the spline.
        value = spline(eventdata)

        return value

    def _is_cached(self, tdm, fitparams_hash):
        """Checks if the ratio and gradients for the given set of fit parameters
        are already cached.
        """
        if((self._cache_fitparams_hash == fitparams_hash) and
           (len(self._cache_ratio) == tdm.n_selected_events)
          ):
            return True
        return False

    def _calculate_ratio_and_gradients(self, tdm, fitparams, fitparams_hash):
        """Calculates the ratio values and ratio gradients for all the events
        given the fit parameters. It caches the results.
        """
        get_data = tdm.get_data

        # Create a 2D event data array holding only the needed event data fields
        # for the PDF ratio spline evaluation.
        eventdata = np.vstack([get_data(fn) for fn in self._data_field_names]).T

        (ratio, gradients) = self._interpolmethod_instance.get_value_and_gradients(
            tdm, eventdata, fitparams)
        # The interpolation works on the logarithm of the ratio spline, hence
        # we need to transform it using the exp function, and we need to account
        # for the exp function in the gradients.
        ratio = np.exp(ratio)
        gradients = ratio * gradients

        # Cache the value and the gradients.
        self._cache_fitparams_hash = fitparams_hash
        self._cache_ratio = ratio
        self._cache_gradients = gradients

    def get_ratio(self, tdm, fitparams, tl=None):
        """Retrieves the PDF ratio values for each given trial event data, given
        the given set of fit parameters. This method is called during the
        likelihood maximization process.
        For computational efficiency reasons, the gradients are calculated as
        well and will be cached.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial event data for which
            the PDF ratio values should get calculated.
        fitparams : dict
            The dictionary with the fit parameter values.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        ratio : 1d ndarray of float
            The PDF ratio value for each given event.
        """
        fitparams_hash = make_params_hash(fitparams)

        # Check if the ratio value is already cached.
        if(self._is_cached(tdm, fitparams_hash)):
            return self._cache_ratio

        self._calculate_ratio_and_gradients(tdm, fitparams, fitparams_hash)

        return self._cache_ratio

    def get_gradient(self, tdm, fitparams, fitparam_name):
        """Retrieves the PDF ratio gradient for the pidx'th fit parameter.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the trial event data for which
            the PDF ratio gradient values should get calculated.
        fitparams : dict
            The dictionary with the fit parameter values.
        fitparam_name : str
            The name of the fit parameter for which the gradient should get
            calculated.
        """
        fitparams_hash = make_params_hash(fitparams)

        # Convert the fit parameter name into the local fit parameter index.
        pidx = self.convert_signal_fitparam_name_into_index(fitparam_name)

        # Check if the gradients have been calculated already.
        if(self._is_cached(tdm, fitparams_hash)):
            return self._cache_gradients[pidx]

        # The gradients have not been calculated yet.
        self._calculate_ratio_and_gradients(tdm, fitparams, fitparams_hash)

        return self._cache_gradients[pidx]



# -*- coding: utf-8 -*-

"""Setup the time-dependent analysis. For now this works on a single dataset.
"""

import argparse
import logging
import numpy as np

from skyllh.core.progressbar import ProgressBar

# Classes to define the source hypothesis.
from skyllh.physics.source import PointLikeSource
from skyllh.physics.flux import PowerLawFlux
from skyllh.core.source_hypo_group import SourceHypoGroup
from skyllh.core.source_hypothesis import SourceHypoGroupManager

# Classes to define the fit parameters.
from skyllh.core.parameters import (
    SingleSourceFitParameterMapper,
    FitParameter
)

# Classes for the minimizer.
from skyllh.core.minimizer import Minimizer, LBFGSMinimizerImpl
from skyllh.core.minimizers.iminuit import IMinuitMinimizerImpl

# Classes for utility functionality.
from skyllh.core.config import CFG
from skyllh.core.random import RandomStateService
from skyllh.core.optimize import SpatialBoxEventSelectionMethod
from skyllh.core.smoothing import BlockSmoothingFilter
from skyllh.core.timing import TimeLord
from skyllh.core.trialdata import TrialDataManager

# Classes for defining the analysis.
from skyllh.core.test_statistic import TestStatisticWilks
from skyllh.core.analysis import (
    TimeIntegratedMultiDatasetSingleSourceAnalysis,
)

# Classes to define the background generation.
from skyllh.core.scrambling import DataScrambler
from skyllh.i3.scrambling import I3SeasonalVariationTimeScramblingMethod
from skyllh.i3.background_generation import FixedScrambledExpDataI3BkgGenMethod

# Classes to define the signal and background PDFs.
from skyllh.core.signalpdf import (
    RayleighPSFPointSourceSignalSpatialPDF,
    SignalBoxTimePDF,
    SignalGaussTimePDF
)
from skyllh.core.backgroundpdf import BackgroundUniformTimePDF
from skyllh.i3.backgroundpdf import (
    DataBackgroundI3SpatialPDF
)
from skyllh.core.pdf import TimePDF

# Classes to define the spatial and energy PDF ratios.
from skyllh.core.pdfratio import (
    SpatialSigOverBkgPDFRatio,
    SigOverBkgPDFRatio
)

# Analysis utilities.
from skyllh.core.analysis_utils import (
    pointlikesource_to_data_field_array
)

from skyllh.core.expectation_maximization import em_fit

# Analysis specific classes for working with the public data.
from skyllh.analyses.i3.publicdata_ps.signal_generator import (
    PDTimeDependentSignalGenerator
)
from skyllh.analyses.i3.publicdata_ps.detsigyield import (
    PublicDataPowerLawFluxPointLikeSourceI3DetSigYieldImplMethod
)
from skyllh.analyses.i3.publicdata_ps.signalpdf import (
    PDSignalEnergyPDFSet
)
from skyllh.analyses.i3.publicdata_ps.pdfratio import (
    PDPDFRatio
)
from skyllh.analyses.i3.publicdata_ps.backgroundpdf import (
    PDDataBackgroundI3EnergyPDF
)
from skyllh.analyses.i3.publicdata_ps.utils import (
    create_energy_cut_spline,
)
from skyllh.analyses.i3.publicdata_ps.time_integrated_ps import (
    psi_func,
)


def change_time_pdf(analysis, gauss=None, box=None):
    """Changes the time pdf.

    Parameters
    ----------
    gauss : dict | None
        None or dictionary with {"mu": float, "sigma": float}.
    box : dict | None
        None or dictionary with {"start": float, "end": float}.
    """

    if gauss is None and box is None:
        raise TypeError("Either gauss or box have to be specified as time pdf.")

    grl = analysis._data_list[0].grl
    # redo this in case the background pdf was not calculated before
    time_bkgpdf = BackgroundUniformTimePDF(grl)
    if gauss is not None:
        time_sigpdf = SignalGaussTimePDF(grl, gauss['mu'], gauss['sigma'])
    elif box is not None:
        time_sigpdf = SignalBoxTimePDF(grl, box["start"], box["end"])

    time_pdfratio = SigOverBkgPDFRatio(
        sig_pdf=time_sigpdf,
        bkg_pdf=time_bkgpdf,
        pdf_type=TimePDF
    )

    # the next line seems to make no difference in the llh evaluation. We keep it for consistency
    analysis._llhratio.llhratio_list[0].pdfratio_list[2] = time_pdfratio
    # this line here is relevant for the llh evaluation
    analysis._llhratio.llhratio_list[0]._pdfratioarray._pdfratio_list[2] = time_pdfratio

    #  change detector signal yield with flare livetime in sample (1 / grl_norm in pdf),
    #  rebuild the histograms if it is changed...


def get_energy_spatial_signal_over_background(analysis, fitparams):
    """Returns the signal over background ratio for
    (spatial_signal * energy_signal) / (spatial_background * energy_background).

    Parameter
    ---------
    fitparams : dict
        Dictionary with {"gamma": float} for energy pdf.

    Returns
    -------
    ratio : 1d ndarray
        Product of spatial and energy signal over background pdfs.
    """

    ratio = analysis._llhratio.llhratio_list[0].pdfratio_list[0].get_ratio(analysis._tdm_list[0], fitparams)
    ratio *= analysis._llhratio.llhratio_list[0].pdfratio_list[1].get_ratio(analysis._tdm_list[0], fitparams)

    return ratio


def change_fluxmodel_gamma(analysis, gamma):
    """Set new gamma for the flux model.

    Parameter
    ---------
    gamma : float
        Spectral index for flux model.
    """

    analysis.src_hypo_group_manager.src_hypo_group_list[0].fluxmodel.gamma = gamma


def change_signal_time(analysis, gauss=None, box=None):
    """Change the signal injection to gauss or box.

    Parameters
    ----------
    gauss : dict | None
        None or dictionary {"mu": float, "sigma": float}.
    box : dict | None
        None or dictionary {"start" : float, "end" : float}.
    """

    analysis.sig_generator.set_flare(box=box, gauss=gauss)


def calculate_TS(analysis, em_results, rss):
    """Calculate the best TS value for the expectation maximization gamma scan.

    Parameters
    ----------
    em_results : 1d ndarray of tuples
        Gamma scan result.
    rss : instance of RandomStateService
        The instance of RandomStateService that should be used to generate
        random numbers from.

    Returns
    -------
    float maximized TS value
    tuple(gamma from em scan [float], best fit mean time [float], best fit width [float])
    (float ns, float gamma) fitparams from TS optimization
    """

    max_TS = 0
    best_time = None
    best_fitparams = None
    for index, result in enumerate(em_results):
        change_time_pdf(analysis,  gauss={"mu": result["mu"], "sigma": result["sigma"]})
        (fitparamset, log_lambda_max, fitparam_values, status) = analysis.maximize_llhratio(rss)
        TS = analysis.calculate_test_statistic(log_lambda_max, fitparam_values)
        if TS > max_TS:
            max_TS = TS
            best_time = result
            best_fitparams = fitparam_values

    return max_TS, best_time, best_fitparams


def run_gamma_scan_single_flare(analysis, remove_time=None, gamma_min=1, gamma_max=5, n_gamma=51):
    """Run em for different gammas in the signal energy pdf

    Parameters
    ----------
    remove_time : float
        Time information of event that should be removed.
    gamma_min : float
        Lower bound for gamma scan.
    gamma_max : float
        Upper bound for gamma scan.
    n_gamma : int
        Number of steps for gamma scan.

    Returns
    -------
    array with "gamma", "mu", "sigma", and scaling factor for flare "ns_em"
    """
    dtype = [("gamma", "f8"), ("mu", "f8"), ("sigma", "f8"), ("ns_em", "f8")]
    results = np.empty(n_gamma, dtype=dtype)

    time = analysis._tdm_list[0].get_data("time")

    for index, g in enumerate(np.linspace(gamma_min, gamma_max, n_gamma)):
        ratio = get_energy_spatial_signal_over_background(analysis, {"gamma": g})
        mu, sigma, ns = em_fit(time, ratio, n=1, tol=1.e-200, iter_max=500, weight_thresh=0,
                            initial_width=5000, remove_x=remove_time)
        results[index] = (g, mu[0], sigma[0], ns[0])

    return results


def unblind_flare(analysis, remove_time=None):
    """Run EM on unscrambled data. Similar to the original analysis, remove the alert event.

    Parameters
    ----------
    remove_time : float
        Time information of event that should be removed.
        In the case of the TXS analysis: remove_time=58018.8711856

    Returns
    -------
    array with "gamma", "mu", "sigma", and scaling factor for flare "ns_em"
    """

    # get the original unblinded data
    rss = RandomStateService(seed=1)
    analysis.unblind(rss)
    time_results = run_gamma_scan_single_flare(analysis, remove_time=remove_time)
    return time_results


def create_analysis(
    datasets,
    source,
    gauss=None,
    box=None,
    refplflux_Phi0=1,
    refplflux_E0=1e3,
    refplflux_gamma=2.0,
    ns_seed=100.0,
    ns_min=0.,
    ns_max=1e3,
    gamma_seed=3.0,
    gamma_min=1.,
    gamma_max=5.,
    kde_smoothing=False,
    minimizer_impl="LBFGS",
    cut_sindec=None,
    spl_smooth=None,
    cap_ratio=False,
    compress_data=False,
    keep_data_fields=None,
    optimize_delta_angle=10,
    tl=None,
    ppbar=None
):
    """Creates the Analysis instance for this particular analysis.

    Parameters:
    -----------
    datasets : list of Dataset instances
        The list of Dataset instances, which should be used in the
        analysis.
    source : PointLikeSource instance
        The PointLikeSource instance defining the point source position.
    gauss : None or dictionary with mu, sigma
        None if no Gaussian time pdf. Else dictionary with {"mu": float, "sigma": float} of Gauss
    box : None or dictionary with start, end
        None if no Box shaped time pdf. Else dictionary with {"start": float, "end": float} of box.
    refplflux_Phi0 : float
        The flux normalization to use for the reference power law flux model.
    refplflux_E0 : float
        The reference energy to use for the reference power law flux model.
    refplflux_gamma : float
        The spectral index to use for the reference power law flux model.
    ns_seed : float
        Value to seed the minimizer with for the ns fit.
    ns_min : float
        Lower bound for ns fit.
    ns_max : float
        Upper bound for ns fit.
    gamma_seed : float | None
        Value to seed the minimizer with for the gamma fit. If set to None,
        the refplflux_gamma value will be set as gamma_seed.
    gamma_min : float
        Lower bound for gamma fit.
    gamma_max : float
        Upper bound for gamma fit.
    kde_smoothing : bool
        Apply a KDE-based smoothing to the data-driven background pdf.
        Default: False.
    minimizer_impl : str | "LBFGS"
        Minimizer implementation to be used. Supported options are "LBFGS"
        (L-BFG-S minimizer used from the :mod:`scipy.optimize` module), or
        "minuit" (Minuit minimizer used by the :mod:`iminuit` module).
        Default: "LBFGS".
    cut_sindec : list of float | None
        sin(dec) values at which the energy cut in the southern sky should
        start. If None, np.sin(np.radians([-2, 0, -3, 0, 0])) is used.
    spl_smooth : list of float
        Smoothing parameters for the 1D spline for the energy cut. If None,
        [0., 0.005, 0.05, 0.2, 0.3] is used.
    cap_ratio : bool
        If set to True, the energy PDF ratio will be capped to a finite value
        where no background energy PDF information is available. This will
        ensure that an energy PDF ratio is available for high energies where
        no background is available from the experimental data.
        If kde_smoothing is set to True, cap_ratio should be set to False!
        Default is False.
    compress_data : bool
        Flag if the data should get converted from float64 into float32.
    keep_data_fields : list of str | None
        List of additional data field names that should get kept when loading
        the data.
    optimize_delta_angle : float
        The delta angle in degrees for the event selection optimization methods.
    tl : TimeLord instance | None
        The TimeLord instance to use to time the creation of the analysis.
    ppbar : ProgressBar instance | None
        The instance of ProgressBar for the optional parent progress bar.

    Returns
    -------
    analysis : TimeIntegratedMultiDatasetSingleSourceAnalysis
        The Analysis instance for this analysis.
    """

    if len(datasets) != 1:
        raise RuntimeError(
            'This analysis supports only analyses with only single datasets '
            'at the moment!')

    if gauss is None and box is None:
        raise ValueError("No time pdf specified (box or gauss)")
    if gauss is not None and box is not None:
        raise ValueError(
            "Time PDF cannot be both Gaussian and box shaped. "
            "Please specify only one shape.")

    # Create the minimizer instance.
    if minimizer_impl == "LBFGS":
        minimizer = Minimizer(LBFGSMinimizerImpl())
    elif minimizer_impl == "minuit":
        minimizer = Minimizer(IMinuitMinimizerImpl(ftol=1e-8))
    else:
        raise NameError(
            f"Minimizer implementation `{minimizer_impl}` is not supported "
            "Please use `LBFGS` or `minuit`.")

    # Define the flux model.
    flux_model = PowerLawFlux(
        Phi0=refplflux_Phi0, E0=refplflux_E0, gamma=refplflux_gamma)

    # Define the fit parameter ns.
    fitparam_ns = FitParameter('ns', ns_min, ns_max, ns_seed)

    # Define the gamma fit parameter.
    fitparam_gamma = FitParameter(
        'gamma', valmin=gamma_min, valmax=gamma_max, initial=gamma_seed)

    # Define the detector signal efficiency implementation method for the
    # IceCube detector and this source and flux_model.
    # The sin(dec) binning will be taken by the implementation method
    # automatically from the Dataset instance.
    gamma_grid = fitparam_gamma.as_linear_grid(delta=0.1)
    detsigyield_implmethod = \
        PublicDataPowerLawFluxPointLikeSourceI3DetSigYieldImplMethod(
            gamma_grid)

    # Define the signal generation method.
    #sig_gen_method = PointLikeSourceI3SignalGenerationMethod()
    sig_gen_method = None

    # Create a source hypothesis group manager.
    src_hypo_group_manager = SourceHypoGroupManager(
        SourceHypoGroup(
            source, flux_model, detsigyield_implmethod, sig_gen_method))

    # Create a source fit parameter mapper and define the fit parameters.
    src_fitparam_mapper = SingleSourceFitParameterMapper()
    src_fitparam_mapper.def_fit_parameter(fitparam_gamma)

    # Define the test statistic.
    test_statistic = TestStatisticWilks()

    # Create the Analysis instance.
    analysis = TimeIntegratedMultiDatasetSingleSourceAnalysis(
        src_hypo_group_manager,
        src_fitparam_mapper,
        fitparam_ns,
        test_statistic,
        sig_generator_cls=PDTimeDependentSignalGenerator
    )

    # Define the event selection method for pure optimization purposes.
    # We will use the same method for all datasets.
    event_selection_method = SpatialBoxEventSelectionMethod(
        src_hypo_group_manager, delta_angle=np.deg2rad(optimize_delta_angle))

    # Prepare the spline parameters.
    if cut_sindec is None:
        cut_sindec = np.sin(np.radians([-2, 0, -3, 0, 0]))
    if spl_smooth is None:
        spl_smooth = [0., 0.005, 0.05, 0.2, 0.3]
    if len(spl_smooth) < len(datasets) or len(cut_sindec) < len(datasets):
        raise AssertionError(
            "The length of the spl_smooth and of the cut_sindec must be equal "
            f"to the length of datasets: {len(datasets)}.")

    # Add the data sets to the analysis.
    pbar = ProgressBar(len(datasets), parent=ppbar).start()
    data_list = []
    energy_cut_splines = []
    for idx, ds in enumerate(datasets):
        # Load the data of the data set.
        data = ds.load_and_prepare_data(
            keep_fields=keep_data_fields,
            compress=compress_data,
            tl=tl)
        data_list.append(data)

        # Create a trial data manager and add the required data fields.
        tdm = TrialDataManager()
        tdm.add_source_data_field('src_array',
                                  pointlikesource_to_data_field_array)
        tdm.add_data_field('psi', psi_func)

        sin_dec_binning = ds.get_binning_definition('sin_dec')
        log_energy_binning = ds.get_binning_definition('log_energy')

        # Create the spatial PDF ratio instance for this dataset.
        spatial_sigpdf = RayleighPSFPointSourceSignalSpatialPDF(
            dec_range=np.arcsin(sin_dec_binning.range))
        spatial_bkgpdf = DataBackgroundI3SpatialPDF(
            data.exp, sin_dec_binning)
        spatial_pdfratio = SpatialSigOverBkgPDFRatio(
            spatial_sigpdf, spatial_bkgpdf)

        # Create the energy PDF ratio instance for this dataset.
        energy_sigpdfset = PDSignalEnergyPDFSet(
            ds=ds,
            src_dec=source.dec,
            flux_model=flux_model,
            fitparam_grid_set=gamma_grid,
            ppbar=ppbar
        )
        smoothing_filter = BlockSmoothingFilter(nbins=1)
        energy_bkgpdf = PDDataBackgroundI3EnergyPDF(
            data.exp, log_energy_binning, sin_dec_binning,
            smoothing_filter, kde_smoothing)

        energy_pdfratio = PDPDFRatio(
            sig_pdf_set=energy_sigpdfset,
            bkg_pdf=energy_bkgpdf,
            cap_ratio=cap_ratio
        )

        pdfratios = [spatial_pdfratio, energy_pdfratio]

        # Create the time PDF ratio instance for this dataset.
        if gauss is not None or box is not None:
            time_bkgpdf = BackgroundUniformTimePDF(data.grl)
            if gauss is not None:
                time_sigpdf = SignalGaussTimePDF(
                    data.grl, gauss['mu'], gauss['sigma'])
            elif box is not None:
                time_sigpdf = SignalBoxTimePDF(
                    data.grl, box["start"], box["end"])
            time_pdfratio = SigOverBkgPDFRatio(
                sig_pdf=time_sigpdf,
                bkg_pdf=time_bkgpdf,
                pdf_type=TimePDF
            )
            pdfratios.append(time_pdfratio)

        analysis.add_dataset(
            ds, data, pdfratios, tdm, event_selection_method)

        energy_cut_spline = create_energy_cut_spline(
            ds, data.exp, spl_smooth[idx])
        energy_cut_splines.append(energy_cut_spline)

        pbar.increment()
    pbar.finish()

    analysis.llhratio = analysis.construct_llhratio(minimizer, ppbar=ppbar)

    # Define the data scrambler with its data scrambling method, which is used
    # for background generation.

    # FIXME: Support multiple datasets for the DataScrambler.
    data_scrambler = DataScrambler(I3SeasonalVariationTimeScramblingMethod(data_list[0]))
    # Create background generation method.
    bkg_gen_method = FixedScrambledExpDataI3BkgGenMethod(data_scrambler)

    analysis.bkg_gen_method = bkg_gen_method
    analysis.construct_background_generator()

    analysis.construct_signal_generator(
        llhratio=analysis.llhratio, energy_cut_splines=energy_cut_splines,
        cut_sindec=cut_sindec, box=box, gauss=gauss)

    return analysis



# -*- coding: utf-8 -*-

"""This module contains utility functions for creating and managing
MultiDimGridPDF instances.
"""

import numpy as np
import os

from skyllh.core.binning import BinningDefinition
from skyllh.core.pdf import MultiDimGridPDF
from skyllh.core.signalpdf import SignalMultiDimGridPDF
from skyllh.core.backgroundpdf import BackgroundMultiDimGridPDF


def kde_pdf_sig_spatial_norm_factor_func(pdf, tdm, fitparams, eventdata):
    """This is the standard normalization factor function for the spatial signal
    MultiDimGridPDF, which is created from KDE PDF values.
    It can be used for the ``norm_factor_func`` argument of the
    ``create_MultiDimGridPDF_from_kde_pdf`` function.
    """
    log10_psi_idx = pdf._axes.axis_name_list.index('log10_psi')
    # psi = tdm.get_data('psi')
    # Convert to psi.
    psi = 10**eventdata[:, log10_psi_idx]
    norm = 1. / (2 * np.pi * np.log(10) * psi * np.sin(psi))
    return norm


def kde_pdf_bkg_norm_factor_func(pdf, tdm, fitparams, eventdata):
    """This is the standard normalization factor function for the background
    MultiDimGridPDF, which is created from KDE PDF values.
    It can be used for the ``norm_factor_func`` argument of the
    ``create_MultiDimGridPDF_from_kde_pdf`` function.
    """
    return 1. / (2 * np.pi)


def create_MultiDimGridPDF_from_photosplinetable(
        ds, data, info_key, splinetable_key, norm_factor_func=None,
        kind=None, tl=None):
    """
    Creates a MultiDimGridPDF instance with pdf values taken from photospline pdf,
    a spline interpolation of KDE PDF values stored in a splinetable on disk.

    Parameters
    ----------
    ds : Dataset instance
        The Dataset instance the PDF applies to.
    data : DatasetData instance
        The DatasetData instance that holds the auxiliary data of the data set.
    info_key : str
        The auxiliary data name for the file containing PDF information.
    splinetable_key : str
        The auxiliary data name for the name of the splinetablefile.
    norm_factor_func : callable | None
        The normalization factor function. It must have the following call
        signature:
            __call__(pdf, tdm, fitparams)
    kind : str | None
        The kind of PDF to create. This is either ``'sig'`` for a
        SignalMultiDimGridPDF or ``'bkg'`` for a BackgroundMultiDimGridPDF
        instance. If set to None, a MultiDimGridPDF instance is created.
    tl : TimeLord instance | None
        The optional TimeLord instance to use for measuring timing information.

    Returns
    -------
    pdf : SignalMultiDimGridPDF instance | BackgroundMultiDimGridPDF instance |
          MultiDimGridPDF instance
        The created PDF instance. Depending on the ``kind`` argument, this is
        a SignalMultiDimGridPDF, a BackgroundMultiDimGridPDF, or a
        MultiDimGridPDF instance.
    """

    if(kind is None):
        pdf_type = MultiDimGridPDF
    elif(kind == 'sig'):
        pdf_type = SignalMultiDimGridPDF
    elif(kind == 'bkg'):
        pdf_type = BackgroundMultiDimGridPDF
    else:
        raise ValueError('The kind argument must be None, "sig", or "bkg"! '
                         'Currently it is '+str(kind)+'!')

    # Load the PDF data from the auxilary files.
    num_dict = ds.load_aux_data(info_key, tl=tl)

    kde_pdf_axis_name_map = ds.load_aux_data('KDE_PDF_axis_name_map', tl=tl)
    kde_pdf_axis_name_map_inv = dict(
        [(v, k) for (k, v) in kde_pdf_axis_name_map.items()])
    for var in num_dict['vars']:
        if(var not in kde_pdf_axis_name_map_inv):
            kde_pdf_axis_name_map_inv[var] = var

    if 'bin_centers' in num_dict:
        bin_centers_key = 'bin_centers'
    elif 'bins' in num_dict:
        bin_centers_key = 'bins'
    else:
        raise KeyError(
            "The PDF information file is missing 'bin_centers' or 'bins' key.")

    axis_binnings = [
        BinningDefinition(
            kde_pdf_axis_name_map_inv[var], num_dict[bin_centers_key][idx])
        for (idx, var) in enumerate(num_dict['vars'])
    ]

    # Getting the name of the splinetable file
    splinetable_file_list = ds.get_aux_data_definition(splinetable_key)
    # This is a list with only one element.
    splinetable_file = os.path.join(ds.root_dir, splinetable_file_list[0])

    pdf = pdf_type(
        axis_binnings,
        path_to_pdf_splinetable=splinetable_file,
        pdf_grid_data=None,
        norm_factor_func=norm_factor_func)

    return pdf


def create_MultiDimGridPDF_from_kde_pdf(
        ds, data, numerator_key, denumerator_key=None, norm_factor_func=None,
        kind=None, tl=None):
    """Creates a MultiDimGridPDF instance with pdf values taken from KDE PDF
    values stored in the dataset's auxiliary data.

    Parameters
    ----------
    ds : Dataset instance
        The Dataset instance the PDF applies to.
    data : DatasetData instance
        The DatasetData instance that holds the auxiliary data of the data set.
    numerator_key : str
        The auxiliary data name for the PDF numerator array.
    denumerator_key : str | None
        The auxiliary data name for the PDF denumerator array.
        This can be None, if no denumerator array is required.
    norm_factor_func : callable | None
        The normalization factor function. It must have the following call
        signature:
            __call__(pdf, tdm, fitparams)
    kind : str | None
        The kind of PDF to create. This is either ``'sig'`` for a
        SignalMultiDimGridPDF or ``'bkg'`` for a BackgroundMultiDimGridPDF
        instance. If set to None, a MultiDimGridPDF instance is created.
    tl : TimeLord instance | None
        The optional TimeLord instance to use for measuring timing information.

    Returns
    -------
    pdf : SignalMultiDimGridPDF instance | BackgroundMultiDimGridPDF instance |
          MultiDimGridPDF instance
        The created PDF instance. Depending on the ``kind`` argument, this is
        a SignalMultiDimGridPDF, a BackgroundMultiDimGridPDF, or a
        MultiDimGridPDF instance.
    """
    if(kind is None):
        pdf_type = MultiDimGridPDF
    elif(kind == 'sig'):
        pdf_type = SignalMultiDimGridPDF
    elif(kind == 'bkg'):
        pdf_type = BackgroundMultiDimGridPDF
    else:
        raise ValueError('The kind argument must be None, "sig", or "bkg"! '
                         'Currently it is '+str(kind)+'!')

    # Load the PDF data from the auxilary files.
    num_dict = ds.load_aux_data(numerator_key, tl=tl)

    denum_dict = None
    if(denumerator_key is not None):
        denum_dict = ds.load_aux_data(denumerator_key, tl=tl)

    kde_pdf_axis_name_map = ds.load_aux_data('KDE_PDF_axis_name_map', tl=tl)
    kde_pdf_axis_name_map_inv = dict(
        [(v, k) for (k, v) in kde_pdf_axis_name_map.items()])
    for var in num_dict['vars']:
        if(var not in kde_pdf_axis_name_map_inv):
            kde_pdf_axis_name_map_inv[var] = var

    if 'bin_centers' in num_dict:
        bin_centers_key = 'bin_centers'
    elif 'bins' in num_dict:
        bin_centers_key = 'bins'
    else:
        raise KeyError(
            "The PDF information file is missing 'bin_centers' or 'bins' key.")

    axis_binnings = [
        BinningDefinition(
            kde_pdf_axis_name_map_inv[var], num_dict[bin_centers_key][idx])
        for (idx, var) in enumerate(num_dict['vars'])
    ]

    vals = num_dict['pdf_vals']
    if(denum_dict is not None):
        # A denumerator is required, so we need to divide the numerator pdf
        # values by the denumerator pdf values, by preserving the correct axis
        # order.
        # Construct the slicing selector for the denumerator pdf values array to
        # match the axis order of the numerator pdf values array.
        selector = []
        for var in num_dict['vars']:
            if(var in denum_dict['vars']):
                # The variable is present in both pdf value arrays. So select
                # all values of that dimension.
                selector.append(slice(None, None))
            else:
                # The variable is not present in the normalization pdf value
                # array, so we need to add a dimension for that variable.
                selector.append(np.newaxis)

        denum = denum_dict['pdf_vals'][tuple(selector)]
        denum_nonzero_mask = denum != 0
        out = np.zeros_like(vals)
        np.divide(vals, denum, where=denum_nonzero_mask, out=out)
        vals = out

    # Set infinite values to NaN.
    vals[np.isinf(vals)] = np.nan
    # Set NaN values to 0.
    vals[np.isnan(vals)] = 0

    pdf = pdf_type(
        axis_binnings,
        path_to_pdf_splinetable=None,
        pdf_grid_data=vals,
        norm_factor_func=norm_factor_func)

    return pdf



import numpy as np
from scipy.stats import norm

from skyllh.core.analysis import TimeIntegratedMultiDatasetSingleSourceAnalysis
from skyllh.core.backgroundpdf import BackgroundUniformTimePDF
from skyllh.core.pdf import TimePDF
from skyllh.core.pdfratio import SigOverBkgPDFRatio
from skyllh.core.random import RandomStateService
from skyllh.core.signalpdf import (
    SignalBoxTimePDF,
    SignalGaussTimePDF,
)


def expectation_em(ns, mu, sigma, t, sob):
    """
    Expectation step of expectation maximization.

    Parameters
    ----------
    ns : float | 1d ndarray of float
        The number of signal neutrinos, as weight for the gaussian flare.
    mu : float | 1d ndarray of float
        The mean time of the gaussian flare.
    sigma: float | 1d ndarray of float
        Sigma of the gaussian flare.
    t : 1d ndarray of float
        Times of the events.
    sob : 1d ndarray of float
        The signal over background values of events, or weights of events

    Returns
    -------
    expectation : list of 1d ndarray of float
        Weighted "responsibility" function of each event to belong to the flare.
    sum_log_denom : float
        Sum of log of denominators.
    """
    ns = np.atleast_1d(ns)
    mu = np.atleast_1d(mu)
    sigma = np.atleast_1d(sigma)

    b_term = (1 - np.cos(10 / 180 * np.pi)) / 2
    N = len(t)
    e_sig = []
    for i in range(len(ns)):
        e_sig.append(norm(loc=mu[i], scale=sigma[i]).pdf(t) * sob * ns[i])
    e_bg = (N - np.sum(ns)) / (np.max(t) - np.min(t)) / b_term
    denom = sum(e_sig) + e_bg

    return [e / denom for e in e_sig], np.sum(np.log(denom))


def maximization_em(e_sig, t):
    """
    Maximization step of expectation maximization.

    Parameters
    ----------
    e_sig : list of 1d ndarray of float
        The weights for each event from the expectation step.
    t : 1d ndarray of float
        The times of each event.

    Returns
    -------
    mu : list of float
        Best fit mean time of the gaussian flare.
    sigma : list of float
        Best fit sigma of the gaussian flare.
    ns : list of float
        Best fit number of signal neutrinos, as weight for the gaussian flare.
    """
    mu = []
    sigma = []
    ns = []
    for i in range(len(e_sig)):
        mu.append(np.average(t, weights=e_sig[i]))
        sigma.append(np.sqrt(np.average(np.square(t - mu[i]), weights=e_sig[i])))
        ns.append(np.sum(e_sig[i]))
    sigma = [max(1, s) for s in sigma]

    return mu, sigma, ns


def em_fit(x, weights, n=1, tol=1.e-200, iter_max=500, weight_thresh=0, initial_width=5000,
            remove_x=None):
    """Run expectation maximization.
    
    Parameters
    ----------
    x : array[float]
        Quantity to run EM on (e.g. the time if EM should find time flares)
    weights :
        weights for each event (e.g. the signal over background ratio)
    fitparams : dict
        Dictionary with value for gamma, e.g. {'gamma': 2}.
    n : int
        How many Gaussians flares we are looking for.
    tol : float
        the stopping criteria for expectation maximization. This is the difference in the normalized likelihood over the
        last 20 iterations.
    iter_max : int
        The maximum number of iterations, even if stopping criteria tolerance (`tol`) is not yet reached.
    weight_thresh : float
        Set a minimum threshold for event weights. Events with smaller weights will be removed.
    initial_width : float
        Starting width for the gaussian flare in days.
    remove_x : float | None
        Specific x of event that should be removed.
    
    Returns
    -------
    Mean, width, normalization factor
    """

    if weight_thresh > 0: # remove events below threshold
        for i in range(len(weights)):
            mask = weights > weight_thresh
            weights[i] = weights[i][mask]
            x[i] = x[i][mask]

    # in case, remove event
    if remove_x is not None:
        mask = x == remove_x
        weights = weights[~mask]
        x = x[~mask]

    # expectation maximization
    mu = np.linspace(x[0], x[-1], n+2)[1:-1]
    sigma = np.ones(n) * initial_width
    ns = np.ones(n) * 10
    llh_diff = 100
    llh_old = 0
    llh_diff_list = [100] * 20

    iteration = 0

    while iteration < iter_max and llh_diff > tol: # run until convergence or maximum number of iterations
        iteration += 1

        e, logllh = expectation_em(ns, mu, sigma, x, weights)

        llh_new = np.sum(logllh)
        tmp_diff = np.abs(llh_old - llh_new) / llh_new
        llh_diff_list = llh_diff_list[:-1]
        llh_diff_list.insert(0, tmp_diff)
        llh_diff = np.max(llh_diff_list)
        llh_old = llh_new
        mu, sigma, ns = maximization_em(e, x)

    return mu, sigma, ns



# -*- coding: utf-8 -*-

import numpy as np

from skyllh.core.binning import UsesBinning
from skyllh.core.pdf import (
    EnergyPDF,
    PDFAxis
)
from skyllh.core.smoothing import (
    UNSMOOTH_AXIS,
    SmoothingFilter,
    HistSmoothingMethod,
    NoHistSmoothingMethod,
    NeighboringBinHistSmoothingMethod
)
from skyllh.core.timing import TaskTimer


class I3EnergyPDF(EnergyPDF, UsesBinning):
    """This is the base class for all IceCube specific energy PDF models.
    IceCube energy PDFs depend soley on the energy and the
    zenith angle, and hence, on the declination of the event.

    The IceCube energy PDF is modeled as a 1d histogram in energy,
    but for different sin(declination) bins, hence, stored as a 2d histogram.
    """
    def __init__(self, data_logE, data_sinDec, data_mcweight, data_physicsweight,
                 logE_binning, sinDec_binning, smoothing_filter):
        """Creates a new IceCube energy PDF object.

        Parameters
        ----------
        data_logE : 1d ndarray
            The array holding the log10(E) values of the events.
        data_sinDec : 1d ndarray
            The array holding the sin(dec) values of the events.
        data_mcweight : 1d ndarray
            The array holding the monte-carlo weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        data_physicsweight : 1d ndarray
            The array holding the physics weights of the events.
            The final data weight will be the product of data_mcweight and
            data_physicsweight.
        logE_binning : BinningDefinition
            The binning definition for the log(E) axis.
        sinDec_binning : BinningDefinition
            The binning definition for the sin(declination) axis.
        smoothing_filter : SmoothingFilter instance | None
            The smoothing filter to use for smoothing the energy histogram.
            If None, no smoothing will be applied.
        """
        super(I3EnergyPDF, self).__init__()

        #self.logger = logging.getLogger(__name__)

        # Define the PDF axes.
        self.add_axis(PDFAxis(name='log_energy',
                              vmin=logE_binning.lower_edge,
                              vmax=logE_binning.upper_edge))
        self.add_axis(PDFAxis(name='sin_dec',
                              vmin=sinDec_binning.lower_edge,
                              vmax=sinDec_binning.upper_edge))

        self.add_binning(logE_binning, 'log_energy')
        self.add_binning(sinDec_binning, 'sin_dec')

        # Create the smoothing method instance tailored to the energy PDF.
        # We will smooth only the first axis (logE).
        if((smoothing_filter is not None) and
           (not isinstance(smoothing_filter, SmoothingFilter))):
            raise TypeError('The smoothing_filter argument must be None or an instance of SmoothingFilter!')
        if(smoothing_filter is None):
            self.hist_smoothing_method = NoHistSmoothingMethod()
        else:
            self.hist_smoothing_method = NeighboringBinHistSmoothingMethod(
                (smoothing_filter.axis_kernel_array, UNSMOOTH_AXIS))

        # We have to figure out, which histogram bins are zero due to no
        # monte-carlo coverage, and which due to zero physics model
        # contribution.

        # Create a 2D histogram with only the MC events to determine the MC
        # coverage.
        (h, bins_logE, bins_sinDec) = np.histogram2d(data_logE, data_sinDec,
            bins = [logE_binning.binedges, sinDec_binning.binedges],
            range = [logE_binning.range, sinDec_binning.range],
            normed = False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered = h > 0

        # Select the events which have MC coverage but zero physics
        # contribution, i.e. the physics model predicts zero contribution.
        mask = data_physicsweight == 0.

        # Create a 2D histogram with only the MC events that have zero physics
        # contribution. Note: By construction the zero physics contribution bins
        # are a subset of the MC covered bins.
        (h, bins_logE, bins_sinDec) = np.histogram2d(data_logE[mask], data_sinDec[mask],
            bins = [logE_binning.binedges, sinDec_binning.binedges],
            range = [logE_binning.range, sinDec_binning.range],
            normed = False)
        h = self._hist_smoothing_method.smooth(h)
        self._hist_mask_mc_covered_zero_physics = h > 0

        # Create a 2D histogram with only the data which has physics
        # contribution. We will do the normalization along the logE
        # axis manually.
        data_weights = data_mcweight[~mask] * data_physicsweight[~mask]
        (h, bins_logE, bins_sinDec) = np.histogram2d(data_logE[~mask], data_sinDec[~mask],
            bins = [logE_binning.binedges, sinDec_binning.binedges],
            weights = data_weights,
            range = [logE_binning.range, sinDec_binning.range],
            normed = False)

        # Calculate the normalization for each logE bin. Hence we need to sum
        # over the logE bins (axis 0) for each sin(dec) bin and need to divide
        # by the logE bin widths along the sin(dec) bins. The result array norm
        # is a 2D array of the same shape as h.
        norms = np.sum(h, axis=(0,))[np.newaxis,...] * np.diff(logE_binning.binedges)[...,np.newaxis]
        h /= norms
        h = self._hist_smoothing_method.smooth(h)

        self._hist_logE_sinDec = h

    @property
    def hist_smoothing_method(self):
        """The HistSmoothingMethod instance defining the smoothing filter of the
        energy PDF histogram.
        """
        return self._hist_smoothing_method
    @hist_smoothing_method.setter
    def hist_smoothing_method(self, method):
        if(not isinstance(method, HistSmoothingMethod)):
            raise TypeError('The hist_smoothing_method property must be an instance of HistSmoothingMethod!')
        self._hist_smoothing_method = method

    @property
    def hist(self):
        """(read-only) The 2D logE-sinDec histogram array.
        """
        return self._hist_logE_sinDec

    @property
    def hist_mask_mc_covered(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage.
        """
        return self._hist_mask_mc_covered

    @property
    def hist_mask_mc_covered_zero_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage but zero physics
        contribution.
        """
        return self._hist_mask_mc_covered_zero_physics

    @property
    def hist_mask_mc_covered_with_physics(self):
        """(read-only) The boolean ndarray holding the mask of the 2D histogram
        bins for which there is monte-carlo coverage and has physics
        contribution.
        """
        return self._hist_mask_mc_covered & ~self._hist_mask_mc_covered_zero_physics

    def assert_is_valid_for_exp_data(self, data_exp):
        """Checks if this energy PDF is valid for all the given experimental
        data.
        It checks if all the data is within the logE and sin(dec) binning range.

        Parameters
        ----------
        data_exp : numpy record ndarray
            The array holding the experimental data. The following data fields
            must exist:

            - 'log_energy' : float
                The logarithm of the energy value of the data event.
            - 'dec' : float
                The declination of the data event.

        Raises
        ------
        ValueError
            If some of the data is outside the logE or sin(dec) binning range.
        """
        logE_binning = self.get_binning('log_energy')
        sinDec_binning = self.get_binning('sin_dec')

        exp_logE = data_exp['log_energy']
        exp_sinDec = np.sin(data_exp['dec'])

        # Check if all the data is within the binning range.
        #if(logE_binning.any_data_out_of_binning_range(exp_logE)):
            #self.logger.warning('Some data is outside the logE range (%.3f, %.3f)', logE_binning.lower_edge, logE_binning.upper_edge)
        #if(sinDec_binning.any_data_out_of_binning_range(exp_sinDec)):
            #self.logger.warning('Some data is outside the sin(dec) range (%.3f, %.3f)', sinDec_binning.lower_edge, sinDec_binning.upper_edge)

    def get_prob(self, tdm, fitparams=None, tl=None):
        """Calculates the energy probability (in logE) of each event.

        Parameters
        ----------
        tdm : instance of TrialDataManager
            The TrialDataManager instance holding the data events for which the
            probability should be calculated for. The following data fields must
            exist:

            - 'log_energy' : float
                The logarithm of the energy value of the event.
            - 'sin_dec' : float
                The sin(declination) value of the event.

        fitparams : None
            Unused interface parameter.
        tl : TimeLord instance | None
            The optional TimeLord instance that should be used to measure
            timing information.

        Returns
        -------
        prob : 1D (N_events,) shaped ndarray
            The array with the energy probability for each event.
        """
        get_data = tdm.get_data

        logE_binning = self.get_binning('log_energy')
        sinDec_binning = self.get_binning('sin_dec')

        logE_idx = np.digitize(
            get_data('log_energy'), logE_binning.binedges) - 1
        sinDec_idx = np.digitize(
            get_data('sin_dec'), sinDec_binning.binedges) - 1

        with TaskTimer(tl, 'Evaluating logE-sinDec histogram.'):
            prob = self._hist_logE_sinDec[(logE_idx,sinDec_idx)]

        return prob
